/*
 * vboot.c -- boot a VS-1000 controller box
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <poll.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include "vs1000.h"

#define DEVICE_NAME "/dev/ttyVC"
#define BOOT_FILE "/etc/vs1000.bin"

#define START_OFFSET	0x00010000UL
#define GO_OFFSET	0x10000000UL

/*
 * Hard-coded ethernet address of the box I have to play with.  We'll
 * fix this for real later...
 *  00:C0:4E:02:00:4F
 */
static unsigned char box_address[6] = { 0x00, 0xC0, 0x4E, 0x02, 0x00, 0x4F };

void config_box(int fd)
{
	struct vs1000_box_config cfg;
	int ret;

	memset(&cfg, 0, sizeof(cfg));
	cfg.boxid = 0;
	cfg.num_ports = 16;
	memcpy(cfg.ether_addr, box_address, 6);
	strcpy(cfg.interface, "eth0");
	ret = ioctl(fd, VS1000_SET_BOX_CONFIG, &cfg);
	if (ret) {
		perror("VS1000_SET_BOX_CONFIG ioctl");
		exit(1);
	}		
}

void send_reset(int fd)
{
	struct vs1000_load_struct vload;
	int	ret;

	vload.boxid = 0;
	vload.offset = 0;
	vload.size = 0xFFFFFFFF;
	vload.buf = 0;
	ret = ioctl(fd, VS1000_LOAD_CODE, &vload);
	if (ret) {
		perror("VS1000_LOAD_CODE ioctl: reset cmd");
		exit(1);
	}
}

void send_query(int fd, int flags)
{
	struct vs1000_load_struct vload;
	int	ret;

	vload.boxid = 0;
	vload.offset = flags;
	vload.size = 0xFFFFFFFE;
	vload.buf = 0;
	ret = ioctl(fd, VS1000_LOAD_CODE, &vload);
	if (ret) {
		perror("VS1000_LOAD_CODE ioctl: query cmd");
		exit(1);
	}
}

void get_report(int fd)
{
	struct vs1000_box_report rpt;
	int	i = 0;

	printf("Getting report from kernel....\n");
	while (1) {
		rpt.flags = i++;
		if (ioctl(fd, VS1000_GET_BOX_REPORT, &rpt))
			break;
		printf("%d\t%02x\t%02x:%02x:%02x:%02x:%02x:%02x\t%s\n",
		       i, rpt.product_id,
		       rpt.ether_addr[0], rpt.ether_addr[1],
		       rpt.ether_addr[2], rpt.ether_addr[3],
		       rpt.ether_addr[4], rpt.ether_addr[5], 
		       (rpt.flags & 1) ? "active" : "loading");
	}
	printf("\n");
}

void load_file(int fd, char *boot_file)
{
	char	buf[1000];
	int	bfd;
	struct vs1000_load_struct vload;
	int	size, ret;

	bfd = open(boot_file, O_RDONLY);
	if (bfd < 0) {
		perror(boot_file);
		exit(1);
	}
	vload.boxid = 0;
	vload.offset = START_OFFSET;
	vload.buf = buf;
	while (1) {
		size = read(bfd, buf, sizeof(buf));
		if (size == 0)
			break;
		if (size < 0) {
			if (errno == EINTR)
				continue;
			perror("read");
			exit(1);
		}
		vload.size = size;
		ret = ioctl(fd, VS1000_LOAD_CODE, &vload);
		if (ret) {
			perror("VS1000_LOAD_CODE ioctl");
			exit(1);
		}
		vload.offset += size;
	}
	close(bfd);
}

void send_go(int fd)
{
	struct vs1000_load_struct vload;
	int	ret;

	vload.boxid = 0;
	vload.size = 0;
	vload.offset = GO_OFFSET;
	vload.buf = 0;
	ret = ioctl(fd, VS1000_LOAD_CODE, &vload);
	if (ret) {
		perror("VS1000_LOAD_CODE ioctl: go cmd");
		exit(1);
	}
	return;
}

void service_device(int fd, char *boot_file)
{
	int	nfds, status, ret;
	struct	pollfd 	poll_ctl;

	printf("In service device...\n");
	poll_ctl.fd = fd;
	poll_ctl.events = POLLOUT;
	poll_ctl.revents = 0;
	while (1) {
		if (ret = poll(&poll_ctl, 1, -1) < 0) {
			if (errno == EINTR)
				continue;
			perror("poll");
			exit(1);
		}
		if (ioctl(fd, VS1000_GET_PROBLEM_BOXES, &status)) {
			perror("VS1000_GET_PROBLEM_BOXES ioctl");
			exit(1);
		}
		printf("Woke up!  Problem status code %d\n", status);
		if (status & 1) {
			printf("Sending reset....\n");
			send_reset(fd);
			sleep(5);
			printf("Sending query (loading only)....\n");
			send_query(fd, 1);
			sleep(1);
			get_report(fd);
			printf("Loading file...\n");
			load_file(fd, boot_file);
			printf("Sending go cmd...\n");
			send_go(fd);
			sleep(2);
			printf("Sending query....\n");
			send_query(fd, 0);
			sleep(1);
			get_report(fd);
		}
	}
}


int main(int argc, char **argv)
{
	int	fd;
	char	*device = DEVICE_NAME;
	char	*boot_file = BOOT_FILE;

	fd = open(device, O_RDONLY);
	if (fd < 0) {
		perror(device);
		exit(1);
	}
	printf("Configuring box...\n");
	config_box(fd);
	printf("Sending query (loading only)....\n");
	send_query(fd, 1);
	sleep(1);
	get_report(fd);
	printf("Sending reset....\n");
	send_reset(fd);
	sleep(5);
	printf("Sending query (loading only)....\n");
	send_query(fd, 1);
	sleep(1);
	get_report(fd);
	printf("Loading file...\n");
	load_file(fd, boot_file);
	printf("Sending go cmd...\n");
	send_go(fd);
	sleep(2);
	printf("Sending query....\n");
	send_query(fd, 0);
	sleep(1);
	get_report(fd);
	service_device(fd, boot_file);
	close(fd);
	return (0);
}

	
	
