/*
 *
 * cmtrl_net.c --- the networking layer for interfacing with the
 * 	Comtrol VS1000 series of products.
 */

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/module.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/mm.h>
#if (LINUX_VERSION_CODE >= 131343) /* 2.1.15 -- XX get correct version */
#include <linux/poll.h>
#endif
#include <linux/in.h>
#include <linux/errno.h>
#include <linux/netdevice.h>
#if (LINUX_VERSION_CODE >= 131343) /* 2.1.15 -- XX get correct version */
#include <linux/inetdevice.h>
#endif
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/stat.h>
#include <linux/timer.h>
#if (LINUX_VERSION_CODE >= 131343) /* 2.1.15 -- XX get correct version */
#include <linux/init.h>
#else
#define __initfunc(x)	x
#endif
#include <linux/tty.h>
#include <linux/tty_flip.h>

#include "vs1000_int.h"
#include "vs1000.h"
#include "version.h"

#undef NETROCKET_DEBUG_HDLC
#undef NETROCKET_DEBUG_RETRANS
#undef NETROCKET_DEBUG_TIMER
#undef NETROCKET_DEBUG_KEEPALIVE
#undef NETROCKET_DEBUG_IN_CMDS
#undef NETROCKET_DEBUG_OUT_CMDS
#undef NETROCKET_DEBUG_OPEN
#undef NETROCKET_DEBUG_INTR
#undef NETROCKET_DEBUG_WRITE
#undef NETROCKET_DEBUG_FLOW
#undef NETROCKET_DEBUG_THROTTLE
#undef NETROCKET_DEBUG_WAIT_UNTIL_SENT
#undef NETROCKET_DEBUG_RECEIVE
#undef NETROCKET_DEBUG_HANGUP
#undef NETROCKET_DEBUG_NETMEM

/*
 * The number of VS1000 boxes we support at once
 */	
#define NUM_VS1000	4

/*
 * This array is used to map concentrator ID's to the struct
 * vs1000_state.  
 */
struct vs1000_state *conc_state[NUM_VS1000];

/*
 * Driver local variables
 */
static struct tty_driver netrocket_driver, callout_driver;
static struct tty_struct *netrocket_table[MAX_NRP_PORTS];
static struct termios *netrocket_termios[MAX_NRP_PORTS];
static struct termios *netrocket_termios_locked[MAX_NRP_PORTS];
static struct nr_port *nrp_table[MAX_NRP_PORTS];
static int netrocket_refcount = 0;
static int nrp_num_ports_open = 0;
static int control_major = 0;

static const char *vs_ctl_name = "VS1000ctl";

static struct wait_queue *wake_daemon_wait;


/* Used to report to the user the vs1000 boxes found on the network */
#define MAX_BOX_REPORTS	32
static struct vs1000_box_report vs1000_box_rpts[MAX_BOX_REPORTS];
static int num_box_rpts;	/* number of valid reports */
static int rpt_loading_only;	/* if non-zero only report available boxes */
	
/*
 * Function declarations
 */
static void hdlc_resync(struct vs1000_state *vs);
static void hdlc_send_ack(struct vs1000_state *vs, __u8 flags);
static void vs1000_set_timer(struct vs1000_state *vs, int ticks);
static int vs_ctl_ioctl(struct inode * inode, struct file * file,
			unsigned int cmd, unsigned long arg);

static void nrp_wait_until_sent(struct tty_struct *tty, int timeout);
static void nrp_flush_buffer(struct tty_struct *tty);
static void nrp_start(struct tty_struct *tty);
static void configure_nr_port(struct nr_port *info);
static void send_recv_buf_to_ldisc(struct nr_port *info);
static void push_chars_to_recv_buf(struct nr_port *info,
				   unsigned char *cp, int len);
static void vs1000_timer(unsigned long arg);
static int init_vs1000_state(struct vs1000_box_config *in, int from_user);
static int free_vs1000_box(int boxid);

#ifdef NETROCKET_DEBUG_NETMEM
extern void show_net_buffers(void);
#endif

/*--------------------------------------------------------
 * Simple inline functions
 */

/*
 * Tell us how many characters are pending in the remote that haven't
 * yet been acknowledged as being transmited.
 */
static inline int get_tx_cnt_remote(struct vs1000_port_state *pstate)
{
	__u16 Get, Put;

	Get = pstate->nGetRemote;
	Put = pstate->nPutRemote;

	if (Put >= Get)
		return (Put - Get);
	else
		return ((__u16) (Put + (~Get) + 1));
}

/*
 * Return how much space is left in the remote's buffer.  We assume
 * the remote has space for 1999 characters
 */
static inline int get_tx_space_remote(struct vs1000_port_state *pstate)
{
	int	c;

	c = 1999 - get_tx_cnt_remote(pstate);
	if (c < 0)
		c = 0;
	return c;
}

static void inline wake_tty_write(struct tty_struct *tty)
{
	wake_up_interruptible(&tty->write_wait);
	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
	    tty->ldisc.write_wakeup)
		(tty->ldisc.write_wakeup)(tty);
}

/* -----------------------------------------------------------------
 * Provide backwards compatibility for kernels prior to 2.1.8.
 */
#if (LINUX_VERSION_CODE < 131336)
int copy_from_user(void *to, const void *from_user, unsigned long len)
{
	int	error;

	error = verify_area(VERIFY_READ, from_user, len);
	if (error)
		return len;
	memcpy_fromfs(to, from_user, len);
	return 0;
}

int copy_to_user(void *to_user, const void *from, unsigned long len)
{
	int	error;
	
	error = verify_area(VERIFY_WRITE, to_user, len);
	if (error)
		return len;
	memcpy_tofs(to_user, from, len);
	return 0;
}

static inline int signal_pending(struct task_struct *p)
{
	return (p->signal & (~p->blocked != 0));
}
#else
#include <asm/uaccess.h>
#endif

static void xmit_packet(struct sk_buff *skb, int pri)
{
#if (LINUX_VERSION_CODE < 131336)
	skb->arp = 1;
	dev_queue_xmit(skb, skb->dev, pri);
#else
	dev_queue_xmit(skb);
#endif
}

#if 0
static void debug_destructor(struct sk_buff * skb)
{
	printk("freeing skb (%d, %d, %d)...", skb->cb[1],
	       skb->cb[0], skb->len);
}
#endif

static void free_packet(struct sk_buff *skb)
{
#if (LINUX_VERSION_CODE < 131336)
	skb->free = 1;
	kfree_skb(skb, 0);
#else
	kfree_skb(skb);
#endif
}

/*--------------------------------------------------------
 * Beginning of the networking code
 */

/*
 * This allocates Linux networking buffer for sending a packet.  Note
 * that the Linux layer takes care of creating the link-level
 * (ethernet) header (i.e., the first 14 bytes of the packet).  We
 * allocate 3 bytes of "network layer header" (struct comtrolhdr),
 * followed by 3 bytes of "transport layer header" which is either the
 * 3 byte header used by the Admin packet, or the 3 byte header used
 * by the HDLC layer.
 */
static struct sk_buff *prepare_packet_hdr(struct device *dev, int packet_len)
{
	struct sk_buff *skb;
	int length;

	if (!dev || !(dev->flags & IFF_UP))
		return NULL;	/* Interface not up! */

	/*
	 * 6 bytes for the header, plus the data plus 15 bytes so that
	 * we can align the packet properly.
	 */
	length = dev->hard_header_len + 6 + packet_len + 15;
	
	skb = alloc_skb(length + dev->hard_header_len + 15, GFP_ATOMIC);
	if (!skb)
		return 0;

#if (LINUX_VERSION_CODE < 131336)
	/* Make sure the 2.0 kernel will free the packet */	
	skb->free = 1;
#endif	

	skb->dev = dev;
	skb->protocol = __constant_htons (ETH_P_COMTROL);
	skb_reserve(skb, (dev->hard_header_len+15)&~15);
	skb->h.raw = skb_put(skb, 6); /* For the headers */
	return skb;
}

/*
 * Send a Product ID query packet
 *
 * If the second argument is NULL, then we're just doing a random
 * query looking for boxes.  If the second argument is non-NULL, then
 * we're querying a specific box with the goal of telling it to listen
 * to *US*.  :-)
 */
static int send_query_packet(struct device *dev, struct vs1000_state *vs)
{
	struct sk_buff *skb;
	struct comtrolhdr *chdr;
	struct adminhdr *ahdr;
	struct admin_product_id_request *areq;

	skb = prepare_packet_hdr(dev, sizeof(struct admin_product_id_request));
	if (!skb) {
		printk("vs1000 --- Alloc_skb failed!\n");
		return -ENOMEM;
	}
	chdr = (struct comtrolhdr *) skb->h.raw;
	chdr->product_id = VS1000_ID;
	chdr->conc_id = vs ? vs->conc_id : 0x00;
	chdr->packet_class = ADMIN_CLASS;
	ahdr = (struct adminhdr *) (skb->h.raw + 3);
	ahdr->len_high = 0;
	ahdr->len_low = sizeof(struct admin_product_id_request) + 1;
	ahdr->admin_type = 2;
	areq = (struct admin_product_id_request *)
		skb_put(skb,sizeof(struct admin_product_id_request));
	memcpy(areq->mac_addr, dev->dev_addr, 6);
	if (vs)
		areq->flags = 0; /* You belong to me.... */
	else
		areq->flags = 1; /* Just testing... */
	dev->hard_header(skb,dev,ETH_P_COMTROL,
			 vs ? vs->ether_addr : dev->broadcast,
			 dev->dev_addr,skb->len);
	xmit_packet(skb, 0);
	return 0;
}

/*
 * Send a reset packet to a specific VS-1000 box
 */
static int send_reset_packet(struct vs1000_state *vs)
{
	struct sk_buff *skb;
	struct comtrolhdr *chdr;
	struct adminhdr *ahdr;
	struct admin_reset *arst;

	skb = prepare_packet_hdr(vs->dev, sizeof(struct admin_reset));
	if (!skb) {
		printk("vs1000 --- Alloc_skb failed!\n");
		return -ENOMEM;
	}
	chdr = (struct comtrolhdr *) skb->h.raw;
	chdr->product_id = VS1000_ID;
	chdr->conc_id = 0x00;
	chdr->packet_class = ADMIN_CLASS;
	ahdr = (struct adminhdr *) (skb->h.raw + 3);
	ahdr->len_high = 0;
	ahdr->len_low = sizeof(struct admin_reset) + 1;
	ahdr->admin_type = ADMIN_RESET_TYPE;
	arst = (struct admin_reset *) skb_put(skb,sizeof(struct admin_reset));
	arst->magic = 0x5555;
	vs->dev->hard_header(skb,vs->dev,ETH_P_COMTROL,vs->ether_addr,
			     vs->dev->dev_addr,skb->len);
	xmit_packet(skb, 0);
	vs->state = VS_STATE_RESET;
	return 0;
}

/*
 * This function sends a Secondary Loader command packet to the vs1000
 * device, and waits for a acknowledgement packet.
 */
static int send_loader_packet(struct vs1000_state *vs,
			      __u8 cmd, __u32 offset, int wait_for_ack,
			      int from_user, int bufsize, void *buf )
{
	struct wait_queue wait = { current, NULL };
	struct sk_buff *skb;
	struct comtrolhdr *chdr;
	struct adminhdr *ahdr;
	char *cp;
	int length, cksum, max_retries, retval, i;
	__u32	tmp;

	add_wait_queue(&vs->loader_wait, &wait);
	max_retries = 5;
	while (1) {
		skb = prepare_packet_hdr(vs->dev, bufsize + 10);
		if (!skb) {
			retval = -ENOMEM;
			goto finished;
		}
		chdr = (struct comtrolhdr *) skb->h.raw;
		chdr->product_id = VS1000_ID;
		chdr->conc_id = vs->conc_id;
		chdr->packet_class = ADMIN_CLASS;
		ahdr = (struct adminhdr *) (skb->h.raw + 3);
		length = bufsize + 20;
		ahdr->len_high = length >> 8;
		ahdr->len_low = length & 0xFF;
		ahdr->admin_type = ADMIN_SEC_LOAD_TYPE;

		/*
		 * Now construct the secondary loader packet
		 */
		cksum = 0;
		cp = (char *) skb_put(skb, bufsize + 10);
		*cp++ = '~'; 
		*cp++ = (bufsize + 6) & 0xFF; cksum += (bufsize + 6) & 0xFF;
		*cp++ = (bufsize + 6) >> 8; cksum += (bufsize + 6) >> 8;
		*cp++ = cmd; cksum += cmd;
		*cp++ = 0;		/* high byte of cmd always zero */
		/* Now we insert the offset */
		tmp = offset;
		*cp++ = tmp & 0xFF; cksum += tmp & 0xFF;
		tmp = tmp >> 8;
		*cp++ = tmp & 0xFF; cksum += tmp & 0xFF;
		tmp = tmp >> 8;
		*cp++ = tmp & 0xFF; cksum += tmp & 0xFF;
		tmp = tmp >> 8;
		*cp++ = tmp & 0xFF; cksum += tmp & 0xFF;
		tmp = tmp >> 8;
		if (bufsize) {
			if (from_user) {
				if (copy_from_user(cp, buf, bufsize)) {
					retval = -EFAULT;
					free_packet(skb);
					goto finished;
				}
			} else
				memcpy(cp, buf, bufsize);
			for (i=0; i < bufsize; i++)
				cksum += *cp++;
		}
		*cp = (~cksum) & 0xFF;
		vs->dev->hard_header(skb,vs->dev,ETH_P_COMTROL,vs->ether_addr,
				     vs->dev->dev_addr,skb->len);
		xmit_packet(skb, 0);
		/*
		 * Now we wait for the reply packet
		 */
		if (!wait_for_ack) {
			retval = 0;
			goto finished;
		}
		vs->load_cmd_pending = cmd;
		current->timeout = jiffies + 3*HZ; /* 3 seconds */
		while (1) {
			current->state = TASK_INTERRUPTIBLE;
			if (signal_pending(current)) {
				retval = -EINTR;
				goto finished;
			}
			if (vs->load_cmd_pending == 0) {
				retval = 0;
				goto finished;
			}
			if (current->timeout == 0)
				break;
			schedule();
		}
		if (max_retries-- == 0)  {
			retval = -EPIPE;
			break;
		}
	}
finished:
	current->state = TASK_RUNNING;
	remove_wait_queue(&vs->loader_wait, &wait);
	return retval;
}

/*
 * Function which handles ioctl from the user-level program to do the
 * booting.   (We're handling the booting code from user-mode, since
 * it's a real mess to read files within kernel code.  The only
 * problem with this approach is that the kernel will need to signal
 * the daemon somehow when it detects that the VS1000 box is need of a
 * reset and re-download of its system code.  XXX we'll handle this later.)
 */
static int send_boot_packet(struct vs1000_state *vs, void *arg)
{
	struct vs1000_load_struct load_args;
	unsigned char go_buf[4];
	int	ret;

	if (copy_from_user(&load_args, arg, sizeof(struct vs1000_load_struct)))
		return -EFAULT;

	if (vs == NULL) {
		if ((load_args.boxid < 0) || (load_args.boxid >= NUM_VS1000))
			return -EINVAL;
		vs = conc_state[load_args.boxid];
		if (!vs)
			return -EINVAL;
	}

	if (!vs->dev || !(vs->dev->flags & IFF_UP))
		return -ENETDOWN;

	if (load_args.size == 0) {
		/*
		 * The go command is indicated by a size of zero.
		 * The offset indicates the start address; the high 16
		 * bits indicates the segment, and the low 16 bits
		 * indicates the offset in the segment.
		 */
		go_buf[0] = (load_args.offset >> 16) & 0xFF;
		go_buf[1] = (load_args.offset >> 24) & 0xFF;
		go_buf[2] = load_args.offset & 0xFF;
		go_buf[3] = (load_args.offset >> 8) & 0xFF;
		ret = send_loader_packet(vs, IOCTL_COMMAND,
					 GO_CMD, 0, 0, 4, go_buf);
		if (ret == 0) {
			vs->state = VS_STATE_CONNECTING;
			vs->send_connect_time = jiffies + HZ;
			vs->resend_counter = 0;
			vs1000_set_timer(vs, HZ);
		}
		return ret;			
	}
	if (load_args.size == 0xFFFFFFFF) {
		/*
		 * If the size is 0xffffffff, send a reset packet
		 */
		return send_reset_packet(vs);
	}
	if (load_args.size == 0xFFFFFFFE) {
		/*
		 * If the size is 0xfffffffe, send a query packet
		 */
		rpt_loading_only = load_args.offset;
		num_box_rpts = 0; /* clear the report buffer */
		return send_query_packet(vs->dev, 0);
	}
	return send_loader_packet(vs, UPLOAD_COMMAND,
				  load_args.offset, 1, 1,
				  load_args.size, load_args.buf);
}

/*
 * HDLC output processing
 */

/*
 * hdlc_resync -- reset the hdlc state variables
 */
static void hdlc_resync(struct vs1000_state *vs)
{
	struct sk_buff *pp;

	vs->in_ack_index = 0;
	vs->out_snd_index= 0;
	vs->next_in_index= 0;
	vs->hdlc_state_flags = 0;
	vs->unacked_pkts = 0;

	/* Clear the retransmit queue */
	while ((pp = skb_dequeue(&vs->rtq)))
		free_packet(pp);
}

/*
 * Prepare an skbuff with the HDLC headers
 */
static struct sk_buff *prepare_hdlc_packet(struct vs1000_state *vs,
					   int size, __u8 flags)
{
	struct sk_buff *skb;
	struct comtrolhdr *chdr;
	struct HDLChdr *hdlc_hdr;

	skb = prepare_packet_hdr(vs->dev, size);
	if (!skb)
		return 0;
	chdr = (struct comtrolhdr *) skb->h.raw;
	chdr->product_id = VS1000_ID;
	chdr->conc_id = vs->conc_id;
	chdr->packet_class = VS1000_CLASS;
	hdlc_hdr = (struct HDLChdr *) (skb->h.raw + 3);
	hdlc_hdr->hdlc_control = flags;
	if (flags & CONTROL_UFRAME)
		hdlc_hdr->out_snd_index = 0;
	else
		hdlc_hdr->out_snd_index = vs->out_snd_index;
	hdlc_hdr->next_in_index = vs->next_in_index;
	return skb;
}

/*
 * This function actually sends the HDLC packet
 */
static void send_hdlc_packet(struct vs1000_state *vs,
			    struct sk_buff *skb)
{
	struct HDLChdr *hdlc_hdr;
	struct sk_buff *tmp;

	hdlc_hdr = (struct HDLChdr *) (skb->h.raw + 3);
	vs->dev->hard_header(skb,vs->dev,ETH_P_COMTROL,vs->ether_addr,
			     vs->dev->dev_addr,skb->len);

	/*
	 * Set the receiver sequence number (for acknowledgement purposes)
	 * Note that we've acknowledged all of our received packets.
	 */
	hdlc_hdr->next_in_index = vs->next_in_index;
	vs->unacked_pkts = 0;
	vs->hdlc_state_flags &= ~HDLC_RCV_TIMER_ACTIVE;

	if (hdlc_hdr->hdlc_control & CONTROL_UFRAME) {
		/*
		 * This is an unsequenced frame; just send it out.
		 */
		xmit_packet(skb, 0);
		return;
	}
	
	 /* Set the outgoing sequence number */
	hdlc_hdr->out_snd_index = vs->out_snd_index++;
	
	/*
	 * Enable the retransmission timer, and put the packet on the
	 * retransmission queue.  Finally send the packet!  
	 */
	if (!(vs->hdlc_state_flags & HDLC_SND_TIMER_ACTIVE)) {
		vs->sender_ack_timer = jiffies + 2*HZ;
		vs->hdlc_state_flags |= HDLC_SND_TIMER_ACTIVE;
		vs1000_set_timer(vs, 2*HZ);
	}
	skb_queue_tail(&vs->rtq,skb);
	tmp=skb_clone(skb, GFP_ATOMIC);
	if (tmp) {
		xmit_packet(tmp, 0);
		/* Clear the receiver ack timer */
		vs->hdlc_state_flags &= ~HDLC_RCV_TIMER_ACTIVE;
	}
	return;
}

/*
 * hdlc_send_ack --- send a bare ack packet
 */
static void hdlc_send_ack(struct vs1000_state *vs, __u8 flags)
{
	struct sk_buff *skb;
	unsigned char *cp;

#ifdef NETROCKET_DEBUG_HDLC	
	printk("Sending bare hdlc ack...");
#endif
	skb = prepare_hdlc_packet(vs, 1, flags);
	if (!skb)
		return;
	cp = skb_put(skb, 1);
	*cp = 0;
	send_hdlc_packet(vs, skb);
	vs->hdlc_state_flags &= ~HDLC_RCV_TIMER_ACTIVE;
}

/*
 * hdlc_send_connect --- send a HDLC connect packet
 */
static void hdlc_send_connect(struct vs1000_state *vs)
{
	struct sk_buff *skb;
	unsigned char *cp;

	skb = prepare_hdlc_packet(vs, 2, CONTROL_UFRAME);
	if (!skb)
		return;
	cp = skb_put(skb, 2);
	*cp++ = RK_CONNECT_ASK;
	*cp = 0;
	send_hdlc_packet(vs, skb);
	vs->state = VS_STATE_CONNECTING;
	hdlc_resync(vs);
#ifdef NETROCKET_DEBUG_HDLC	
	printk("Sent HDLC connect!\n");
#endif
}

/*
 * hdlc_retransmit -- retransmit the packets in the retranmission queue.
 */
static void hdlc_retransmit(struct vs1000_state *vs)
{
	struct sk_buff *skb, *tmp, *next;
#ifdef NETROCKET_DEBUG_RETRANS
	struct HDLChdr *hdlc_hdr;
#endif

	skb = skb_peek(&vs->rtq);
	while (skb && skb != (struct sk_buff *)&vs->rtq) {
		next = skb->next;
#if (LINUX_VERSION_CODE >= 131336)
#define SKB_RETRY cb[0]
#else
#define SKB_RETRY tries
#endif
#ifdef NETROCKET_DEBUG_RETRANS
		hdlc_hdr = (struct HDLChdr *) (skb->h.raw + 3);
		printk("retransmitting packet %d (%d times)\n",
		       hdlc_hdr->out_snd_index, skb->SKB_RETRY);
#endif
		if (skb->SKB_RETRY++ > 60) {
			/* This is the last time... */
			skb_unlink(skb);
			tmp = skb;
		} else
			tmp=skb_clone(skb, GFP_ATOMIC);
		if (tmp)
			xmit_packet(tmp, -1);
		skb = next;
	}
	if (skb_peek(&vs->rtq) == NULL) {
#ifdef NETROCKET_DEBUG_HDLC	
		printk("Clearing retransmit flag...");
#endif
		vs->hdlc_state_flags &= ~HDLC_SND_TIMER_ACTIVE;
	}
}

/*
 * Reset the port status
 */
static void port_resync_all(struct vs1000_state *vs)
{
	struct vs1000_port_state	*pstate;
	struct nr_port			*info;
	int				i;

	for (i=0; i < vs->num_ports; i++) {
		pstate = vs->port_state + i;

		pstate->nPutRemote = 0;
		pstate->nGetRemote = 0;
		pstate->nGetLocal = 0;
		pstate->modemStatusRegister = 0;	
		pstate->errorStatusRegister = 0;
		pstate->update_queue = 0;

		info = nrp_table[i + vs->base_port];
		if (info) {
			info->update_baud_rate = 1;
			info->update_control_settings = 1;
			vs->need_to_send = 1;
		}
	}
	if (vs->need_to_send)
		vs1000_timer((unsigned long) vs);
}

/*****************************************************************
 * This section of code handles the transmissions of commands to the
 * remote.
 *
 * The way we handle things is to overload the Linux's skbuff's
 * pkt_type to hold the current port, which is needed to determine
 * whether we need to send a RK_PORT_SET subpacket or not.
 */

/*
 * Send a RK_PORT_SET if necessary.  Make sure we have at least 3
 * bytes in the outgoing packet before calling this routine!  Returns
 * 0 if OK, 1 on error.
 */
static inline int vs_out_set_port(struct sk_buff *skb, int port)
{
	unsigned char *cp;

	if (skb->pkt_type == port)
		return 0;
	if (skb_tailroom(skb) < 3)
		return 1;
	cp = skb_put(skb, 2);
	*cp++ = RK_PORT_SET;
	*cp = port;
	skb->pkt_type = port;
#ifdef NETROCKET_DEBUG_OUT_CMDS
	printk("\nPort %d...", port);
#endif
	return 0;
}

/*
 * Send 2 byte commads; make sure there are at least 6 bytes of free space!
 */
static inline int vs_out_2byte_cmd(struct sk_buff *skb,int port,
				    __u8 cmd, __u16 value)
{
	unsigned char *cp;
	
	if (vs_out_set_port(skb, port))
		return 1;
	if (skb_tailroom(skb) < 3)
		return 1;
	cp = skb_put(skb, 3);
	*cp++ = cmd;
	*cp++ = value & 0xFF;
	*cp = value >> 8;
	return 0;
}

/*
 * Send 4 byte commads; make sure there are at least 8 bytes of free space!
 */
static inline int vs_out_4byte_cmd(struct sk_buff *skb,int port,
				    __u8 cmd, __u16 value)
{
	unsigned char *cp;
	
	if (vs_out_set_port(skb, port))
		return 1;
	if (skb_tailroom(skb) < 5)
		return 1;
	cp = skb_put(skb, 5);
	*cp++ = cmd;
	*cp++ = value & 0xFF;
	*cp++ = value >> 8;
	*cp++ = value >> 16;
	*cp = value >> 24;
	
	return 0;
}

/*
 * Transmit any characters that might be present
 */
static inline int vs_out_xmit(struct sk_buff *skb, struct nr_port *info,
			      struct vs1000_port_state *pstate)
{
	unsigned char	*cp;
	int		c, space, total_len = 0;
	struct tty_struct *tty;
	
	if (!info || (!info->xmit_cnt && !info->x_char) ||
	    !get_tx_space_remote(pstate))
		return 1;
	if (vs_out_set_port(skb, info->line))
		return 1;
	if (skb_tailroom(skb) < 5)
		return 1;
	cp = skb_put(skb, 3);
	*cp++ = RK_DATA_BLK;
	if (info->x_char) {
		*(skb_put(skb, 1)) = info->x_char;
		total_len++;
		pstate->nPutRemote++;
		info->x_char = 0;
	}
	while (1) {
		c = skb_tailroom(skb) - 1;
#if 0
		printk("tailroom=%d...", c);
#endif
		if (c > info->xmit_cnt)
			c = info->xmit_cnt;
		if (c > (XMIT_BUF_SIZE - info->xmit_tail))
			c = XMIT_BUF_SIZE - info->xmit_tail;
		space = get_tx_space_remote(pstate);
		if (c > space)
			c = space;
		if (c <= 0)
			break;
#if 0
		printk("c=%d...", c);
#endif
		memcpy(skb_put(skb, c), info->xmit_buf + info->xmit_tail, c);
		info->xmit_tail += c;
		info->xmit_tail &= XMIT_BUF_SIZE-1;
		info->xmit_cnt -= c;
		total_len += c;
		pstate->nPutRemote += (__u16) c;
	}
#ifdef NETROCKET_DEBUG_OUT_CMDS
	printk("ttyV%d: sent %d bytes...", info->line, total_len);
#endif
	*cp++ = total_len & 0xFF;
	*cp++ = total_len >> 8;

	tty = info->tty;
	if ((info->xmit_cnt < WAKEUP_CHARS) && tty)
		wake_tty_write(tty);
	return 0;
}

/*
 * Figure out what we need to send to the remote.  If this function is
 * called, it's highly likely that something needs to be done....
 */
#define MAX_CMD_PKT 984
static void send_remote_cmds(struct vs1000_state *vs)
{
	int	i;
	struct	sk_buff *skb = NULL;
	unsigned char *cp;
	struct vs1000_port_state *pstate;
	struct nr_port *info;
	int	aborted, retry_start = 0;
	int	send_packet;

retry:
	send_packet = 0;
	vs->need_to_send = 0;
	aborted = 0;
	skb = prepare_hdlc_packet(vs, MAX_CMD_PKT, CONTROL_IFRAME);
	if (!skb)
		return;
	skb->pkt_type = -1;
	
	for (i=retry_start; i < vs->num_ports; i++) {
		retry_start = i;
		pstate = vs->port_state + i;
		if (pstate->update_queue) {
			if (vs_out_2byte_cmd(skb, i, RK_QIN_STATUS,
					     pstate->nGetLocal)) {
				aborted = 1;
				break;
			}
			pstate->update_queue = 0;
			send_packet = 1;
		}
		info = nrp_table[i + vs->base_port];
		if (!info)
			continue;
		if (info->update_baud_rate) {
#ifdef NETROCKET_DEBUG_OUT_CMDS
			printk("set baud rate to %d...", info->baud_rate);
#endif
			if (vs_out_4byte_cmd(skb, i, RK_BAUD_SET,
					     info->baud_rate)) {
				aborted = 1;
				break;
			}
			info->update_baud_rate = 0;
			send_packet = 1;
		}
		if (info->update_control_settings) {
#ifdef NETROCKET_DEBUG_OUT_CMDS
			printk("set control byte to %d...",
			       info->control_settings);
#endif
			if (vs_out_2byte_cmd(skb, i, RK_CONTROL_SET,
					     info->control_settings)) {
				aborted = 1;
				break;
			}
			info->update_control_settings = 0;
			send_packet = 1;
		}
		if (info->update_mcr) {
#ifdef NETROCKET_DEBUG_OUT_CMDS
			printk("set MCR to %d...",
			       info->modem_control_register);
#endif
			if (vs_out_2byte_cmd(skb, i, RK_MCR_SET,
					     info->modem_control_register)) {
				aborted = 1;
				break;
			}
			info->update_mcr = 0;
			send_packet = 1;
		}
		if (info->update_action) {
#ifdef NETROCKET_DEBUG_OUT_CMDS
			printk("set action to %d...",
			       info->action_register);
#endif
			if (vs_out_2byte_cmd(skb, i, RK_ACTION_SET,
					     info->action_register)) {
				aborted = 1;
				break;
			}
			info->update_action = 0;
			info->action_register = 0;
			send_packet = 1;
		}
		/*
		 * Transmit characters in our transmit buffer
		 */
		if (info->x_char ||
		    (info->xmit_cnt &&
		     !info->tty->hw_stopped && !info->tty->stopped)) {
			if (!vs_out_xmit(skb, info, pstate))
				send_packet = 1;
		}
	}
	if (send_packet) {
		cp = skb_put(skb, 1);
		*cp = 0;
		skb->pkt_type = PACKET_HOST;
#ifdef NETROCKET_DEBUG_OUT_CMDS
		printk("packet sent (%d, %d)...", vs->out_snd_index,
		       vs->in_ack_index);
#endif
		send_hdlc_packet(vs, skb);
	} else {
		printk("Freeing unneeded skbuff...");
		free_packet(skb);
	}
	if (aborted)
		goto retry;
}


/*
 * Network reception happens here.
 */

/*
 * This routine processes incoming HDLC packets, and handles the
 * acknowledging them.
 */
static void process_hdlc_packet(struct sk_buff *skb,
				struct vs1000_state *vs)
{
	struct HDLChdr *hdlc_hdr, *hdr2;
	unsigned long flags;
	struct sk_buff	*pp;
	unsigned int	ack_index, ns_save, tmp, port_num;
	unsigned char *cp;
	int	len, cmd, last_cmd, ack_count, space;
	struct nr_port *info;
	struct vs1000_port_state *pstate;

	/*
	 * We only process HDLC packet when we are connected, waiting
	 * for a response from a connect request, or connected but
	 * waiting for the response from the keepalive ping.
	 */
	switch (vs->state) {
	case VS_STATE_CONNECTING:
	case VS_STATE_CONNECTED:
	case VS_STATE_MAYBE_DEAD:
		break;
	default:
		return;
	}

	hdlc_hdr = (struct HDLChdr *) (skb->h.raw + 3);
	skb_pull(skb, sizeof(struct HDLChdr));	
	if (!(hdlc_hdr->hdlc_control & CONTROL_UFRAME)) {
#ifdef NETROCKET_DEBUG_HDLC	
		printk("HDLC packet (%d) #%d...",
		       hdlc_hdr->hdlc_control, hdlc_hdr->out_snd_index);
#endif
		/*
		 * Check for out of order packets.
		 */
		if (hdlc_hdr->out_snd_index != vs->next_in_index) {
#ifdef NETROCKET_DEBUG_HDLC	
			printk("hdlc: out of order packet recv'ed (%d, %d)\n",
			       hdlc_hdr->out_snd_index, vs->next_in_index);
#endif
			hdlc_send_ack(vs, CONTROL_UFRAME);
			return;
		}

		/*
		 * Bump the number of count of unacked packets.  If we
		 * exceed the window size, request an immediate ack
		 * packet.  Otherwise, if the receiver ack timer
		 * hasn't been set yet, turn it on, so that we ack
		 * after an appropriate time.
		 */
		vs->unacked_pkts++;
		if (vs->unacked_pkts > vs->pkt_window_size)
			hdlc_send_ack(vs, CONTROL_UFRAME);
		else if (!(vs->hdlc_state_flags & HDLC_RCV_TIMER_ACTIVE)) {
			vs->receiver_ack_timer = jiffies + HZ;
			vs->hdlc_state_flags |= HDLC_RCV_TIMER_ACTIVE;
			vs1000_set_timer(vs, HZ);
		}
		vs->next_in_index++;
	}

	/*
	 * Check if receiver has acknowledged any of our packets.
	 * If so, prune the retransmission queue.
	 */
	if (vs->in_ack_index != hdlc_hdr->next_in_index) {
		vs->in_ack_index = hdlc_hdr->next_in_index;
		ack_index = vs->in_ack_index - 1;
		ack_count = 0;

		save_flags(flags);
		cli();

		/*
		 * Note that this assumes that we will be removing at
		 * least one item from the queue.  Given that the ack
		 * index has changed, this is a reasonable assumption.
		 */
		while ((pp = skb_dequeue(&vs->rtq))) {
			hdr2 = (struct HDLChdr *) (pp->h.raw + 3);
			ns_save = hdlc_hdr->out_snd_index;

			free_packet(pp);
			ack_count++;

			if (ns_save == ack_index)
				break;  
		}
		if (ack_count) {
			if (vs->state == VS_STATE_MAYBE_DEAD) {
#ifdef NETROCKET_DEBUG_KEEPALIVE
				printk("->VS_STATE_CONNECTED...");
#endif
				vs->state = VS_STATE_CONNECTED;
			}
			vs->keep_alive_timer = jiffies + KEEP_ALIVE_TIME;
			vs1000_set_timer(vs, KEEP_ALIVE_TIME);
		}
		if (!skb_peek(&vs->rtq)) {
#ifdef NETROCKET_DEBUG_HDLC	
			printk("all acked! ");
#endif
			vs->hdlc_state_flags &= ~HDLC_SND_TIMER_ACTIVE;
		} else {
#ifdef NETROCKET_DEBUG_HDLC	
			printk("not all acked! ");
#endif
			vs->sender_ack_timer = jiffies + 2*HZ;
			vs->hdlc_state_flags |= HDLC_SND_TIMER_ACTIVE;
			vs1000_set_timer(vs, HZ);
		}
		restore_flags(flags);
	}

	/*
	 * OK, now we actually process the data in the packet.
	 *
	 * XXX Currently, we assume the packet is null terminated, and
	 * we don't do length checking.  We need to fix this.  Right
	 * now, a malformed packet can cause the system to crash.  If
	 * it's any consolation the NT driver doesn't do this sanity
	 * checking either, but Unix drivers should be more robust
	 * than NT drivers.  Just because it's easy to make NT machines
	 * crash with malformed packets doesn't mean we have to make
	 * Unix machines bug-compatible with NT.    :-)
	 */
	cp = skb->data;
	len = skb->len;
	port_num = -1;
	info = NULL;
	pstate = NULL;
	cmd = 0;
	while ((len > 0) && *cp) {
		last_cmd = cmd;
		cmd = *cp++; len--;
		switch (cmd) {
		case RK_CONNECT_CHECK:
#ifdef NETROCKET_DEBUG_IN_CMDS
			printk("vs1000 connect...");
#endif
			break;	/* Is a no-op on the server */

		case RK_CONNECT_REPLY:
#ifdef NETROCKET_DEBUG_IN_CMDS
			printk("connect reply...");
#endif
			if (vs->state == VS_STATE_CONNECTING) {
				vs->state = VS_STATE_CONNECTED;
				vs->keep_alive_timer = jiffies +
					KEEP_ALIVE_TIME;
				vs1000_set_timer(vs, KEEP_ALIVE_TIME);
			} else {
				printk("Unexpected VS-1000 connect!\n");
			}
			port_resync_all(vs);
			break;

		case RK_CONNECT_ASK:  
			break;		/* should not see this on server */

		case RK_PORT_SET:
			if (len < 1)
				break;
			port_num = *cp++; len--;
#ifdef NETROCKET_DEBUG_IN_CMDS
			printk("port %d...", port_num);
#endif
			if (port_num < vs->num_ports) {
				info = nrp_table[port_num + vs->base_port];
				pstate = vs->port_state + port_num;
			} else {
				/* should never happen */
				printk("vs1000: bad port from server!\n");
				port_num = -1;
				info = NULL;
				pstate = NULL;
			}
			break;

		case RK_MSR_SET:  /* modem status reg pkt */
			if (len < 2)
				break;
			tmp = *cp++;
			tmp += (*cp++) << 8;
			len -= 2;
#ifdef NETROCKET_DEBUG_IN_CMDS
			printk("msr %d...", tmp);
#endif
			if (pstate)
				pstate->modemStatusRegister = tmp;
			if (!info || !info->tty)
				break;
			if ((tmp & MSR_CD_ON) && (!info->cd_status)) {
#if (defined(ROCKET_DEBUG_OPEN) || defined(ROCKET_DEBUG_INTR) || defined(ROCKET_DEBUG_HANGUP))
				printk("ttyR%d CD now on...", info->line);
#endif
				wake_up_interruptible(&info->open_wait);
				info->cd_status = 1;
			} else if (!(tmp & MSR_CD_ON) && (info->cd_status)) {
#if (defined(ROCKET_DEBUG_OPEN) || defined(ROCKET_DEBUG_INTR) || defined(ROCKET_DEBUG_HANGUP))
				printk("ttyR%d CD now off...", info->line);
#endif
				if (!C_CLOCAL(info->tty)) {
#ifdef ROCKET_DEBUG_HANGUP
					printk("CD drop, calling hangup.\n");
#endif
					tty_hangup(info->tty);
				}
			}
			break;

		case RK_ESR_SET:  /* error status reg pkt */
			if (len < 2)
				break;
			tmp = *cp++;
			tmp += (*cp++) << 8;
			len -=2;
#ifdef NETROCKET_DEBUG_IN_CMDS
			printk("esr %d...", tmp);
#endif
			if (pstate)
				pstate->errorStatusRegister = tmp;
			break;

		case RK_ACTION_ACK:  /* reaction response pkt */
			if (len < 2)
				break;
			tmp = *cp++;
			tmp += (*cp++) << 8;
			len -=2;
#ifdef NETROCKET_DEBUG_IN_CMDS
			printk("action resp %d...", tmp);
#endif
			if (pstate)
				pstate->actionResponse = tmp;
			break;

		case RK_QIN_STATUS:  /* qin status report */
			if (len < 2)
				break;
			tmp = *cp++;
			tmp += (*cp++) << 8;
			len -= 2;
			if (!pstate)
				break;
			pstate->nGetRemote = tmp;
			if (!info || !info->tty)
				break;
			space = get_tx_space_remote(pstate);
			if (space > WAKEUP_CHARS) {
				if (info->xmit_cnt)
					vs->need_to_send = 1;
				if (info->xmit_cnt < WAKEUP_CHARS)
					wake_tty_write(info->tty);
			}
#ifdef NETROCKET_DEBUG_IN_CMDS
			printk("ttyV%d: Space %d (%d, %d)\n",
			       port_num, space, pstate->nPutRemote,
			       pstate->nGetRemote);
#endif
			break;

		case RK_DATA_BLK:    /* incoming characters */
			if (len < 2)
				break;
			tmp = *cp++;
			tmp += (*cp++) << 8;
			len -= 2;
#ifdef NETROCKET_DEBUG_IN_CMDS
			printk("data_blk len=%d...", tmp);
#endif
			if (len < tmp)
				tmp = len; /* should never happen */
			if (!info || !info->tty ||
			    (info->flags & VS1000_CLOSING) ||
			    !(info->flags & VS1000_INITIALIZED)) {
				/*
				 * The port's not open, so we just
				 * tell the box that we've processed
				 * the characters
				 */
				if (pstate) {
					cp += tmp; len -= tmp;
					pstate->nGetLocal += tmp; 
					pstate->update_queue = 1;
					vs->need_to_send = 1;
				}
				break;
			}
			/*
			 * If there are characters in the buffer
			 * already, or there isn't enough space to
			 * take the incoming characters, push it to
			 * the receive buffer
			 */
			if (info->recv_cnt ||
			    tmp >= info->tty->ldisc.receive_room(info->tty)) {
				push_chars_to_recv_buf(info, cp, tmp);
				cp += tmp; len -= tmp;
				break;
			}
			/*
			 * Just send it straight to the line discpline.
			 */
			info->tty->ldisc.receive_buf(info->tty, cp, NULL, tmp);
			cp += tmp; len -= tmp;
			pstate->nGetLocal += tmp; 
			pstate->update_queue = 1;
			vs->need_to_send = 1;
			break;
		default:
			printk("vs1000: bad subtype 0x%02x "
			       "(last=0x%02x,len=%d)\n", cmd, 
			       last_cmd, len);
			break;
		}
	}
	if (vs->need_to_send)
			vs1000_set_timer(vs, 0);
}


/*
 * This routine is used to check the ADMIN secondary loader
 * acknowledgement packets.  Note that the IOCTL "go" command is never
 * acknowledged, so don't grow old waiting for it!
 */
static void process_sec_load_ack(struct sk_buff *skb,
				 unsigned char *admin_data,
				 struct vs1000_state *vs)
{
	unsigned char *cp = admin_data;
	int	length, i;
	unsigned int cksum, cmd;

	if (!vs || !vs->load_cmd_pending)
		return;

	if (skb->len < 5) {
		printk("process_sec_load_ack: packet too small (%d)\n",
		       skb->len);
		return;		/* too small */
	}
	if (*cp != '|') {
		printk("process_sec_load_ack: wrong header byte (%d)\n", *cp);
		return;		/* not an ack */
	}
	cp++;
	length = *cp;  cksum = *cp++;
	length += *cp << 8;  cksum += *cp++;
	cmd = *cp; cksum += *cp++;
	cmd += *cp << 8; cksum += *cp++;
	if (cmd != (vs->load_cmd_pending | 0x80)) {
		printk("process_sec_load_ack: wrong command (%d, %d)\n",
		       cmd, vs->load_cmd_pending);
		return;	
	}
	if (length+4 > skb->len) {
		printk("process_sec_load_ack: packet length bad (%d, %d)\n",
		       length, skb->len);
		return;		/* too small */
	}
	for (i=2; i < length; i++) {
		cksum += *cp++;
	}
	cksum = (~cksum) & 0xFF;
	if (cksum != *cp) {
		printk("process_sec_load_ack: incorrect checksum (%d, %d)\n",
		       cksum, *cp);
		return;
	}
	vs->load_cmd_pending = 0;
	wake_up_interruptible(&vs->loader_wait);
}

static struct vs1000_state *find_vs_struct(char *eth_addr)
{
	struct vs1000_state 	*vs;
	int			i;
	
	for (i=0; i < NUM_VS1000; i++) {
		vs = conc_state[i];
		if (!vs)
			continue;
		if (memcmp(vs->ether_addr, eth_addr, 6))
			return vs;
	}
	return 0;
}

static int comtrol_rcv(struct sk_buff *skb, struct device *dev,
		      struct packet_type *pt)
{
	struct	comtrolhdr	*chdr = (struct comtrolhdr *) skb->h.raw;
	struct  ethhdr		*ehdr = skb->mac.ethernet;
	unsigned char *hdr;
	struct adminhdr *ahdr;
	unsigned char *admin_data;
	struct vs1000_state *vs;
	struct vs1000_box_report *rpt;
	
	/*
	 * Drop packets not meant for us (that may show up if the
	 * interface is in promiscuous mode.)
	 */
	if (skb->pkt_type == PACKET_OTHERHOST)
		goto drop;

	hdr = skb_pull(skb, sizeof(struct comtrolhdr));
	if (hdr == NULL)
		goto drop;

	vs = 0;
	if (chdr->conc_id < NUM_VS1000)
		vs = conc_state[chdr->conc_id];

	switch (chdr->packet_class) {
	case ADMIN_CLASS:
		ahdr = (struct adminhdr *) hdr;
		admin_data = skb_pull(skb, sizeof(struct adminhdr));
		switch (ahdr->admin_type) {
		case ADMIN_SEC_LOAD_TYPE:
			process_sec_load_ack(skb, admin_data, vs);
			break;
		case ADMIN_PRD_ID_REP_TYPE:
#if 0
			printk("Response from device %02x, mac addr "
			       "%02x:%02x:%02x:%02x:%02x:%02x %s\n",
			       chdr->product_id, admin_data[0],
			       admin_data[1], admin_data[2],
			       admin_data[3], admin_data[4],
			       admin_data[5], admin_data[6] ?
			       "active" : "loading");
#endif
			/*
			 * note:  admin_data should be same as
			 * ehdr->h_source!
			 */
			vs = find_vs_struct(admin_data);
			if (vs) {
				printk("VS struct found!\n");
				if (vs->state >= VS_STATE_CONNECTING &&
				    admin_data[6] == 0)
					vs->state = VS_STATE_RESET;
			}
			if ((num_box_rpts < MAX_BOX_REPORTS) &&
			    (!rpt_loading_only || !admin_data[6])) {
				rpt = vs1000_box_rpts + num_box_rpts;
				memcpy(rpt->ether_addr, ehdr->h_source, 6);
				rpt->product_id = chdr->product_id;
				rpt->flags = admin_data[6];
				num_box_rpts++;
			}
			break;
		}
		break;
	case VS1000_CLASS:
		if (vs)
			process_hdlc_packet(skb, vs);
		break;
	}
	
drop:
        free_packet(skb);
        return(0);
}


struct packet_type comtrol_packet_type=
{
	0,
	NULL,
	comtrol_rcv,
	NULL,
	NULL
};

/****************************************************************
 * The timer function happens here.
 */

static void vs1000_timer(unsigned long arg)
{
	struct vs1000_state *vs = (struct vs1000_state *) arg;
	int	next_time = T_INFINITE, tmp;
	int	i;

	if (atomic_dec_and_test(&vs->timer_active) == 0) {
		/* The timer must be already running */
		atomic_inc(&vs->timer_active);
		return;
	}

#ifdef NETROCKET_DEBUG_TIMER
	printk("In vs1000_timer (%d)...", atomic_read(&vs->timer_active));
#endif

	/*
	 * If we're in the state VS_STATE_CONNECTING, then send the
	 * HDLC connect packet.  Once we receive the connect packet,
	 * we will enter the VS_STATE_CONNECTED.
	 */
	if (vs->state == VS_STATE_CONNECTING) {
		tmp = vs->send_connect_time - jiffies;
		if (tmp > 0) {
			if (tmp < next_time)
				next_time = tmp;
		} else {
			if (vs->resend_counter >= 5) {
#ifdef NETROCKET_DEBUG_TIMER
				printk("VS-1000 connect timeout");
#endif
				send_reset_packet(vs);
			} else {
				vs->resend_counter++;
				send_query_packet(vs->dev, vs);
				hdlc_send_connect(vs);
				vs->send_connect_time = jiffies + 2*HZ;
				if (HZ < next_time)
					next_time = 2*HZ;
			}
		}
	}
#ifdef NETROCKET_DEBUG_TIMER
	printk("hdlc_state_flags=%d...", vs->hdlc_state_flags);
#endif
	/*
	 * If we sent a keep-alive ping, and don't get an answer back
	 * in time, declare the device dead and send a reset packet.
	 */
	if (vs->state == VS_STATE_MAYBE_DEAD) {
		tmp = vs->keep_alive_timer - jiffies;
		if (tmp > 0) {
			if (tmp < next_time)
				next_time = tmp;
		} else {
#ifdef NETROCKET_DEBUG_KEEPALIVE
			printk("->VS_STATE_DEAD...");
#endif
			vs->state = VS_STATE_DEAD;
			vs->hdlc_state_flags = 0;
			wake_up_interruptible(&wake_daemon_wait);
		}
	}
	if (vs->state == VS_STATE_CONNECTED) {
		/*
		 * Check to see if we need to service the keep alive
		 * timer.  If so, we go into state
		 * VS_STATE_MAYBE_DEAD, and we don't send any other
		 * packets to the device until we're sure we're alive
		 */
		tmp = vs->keep_alive_timer - jiffies;
		if (tmp > 0) {
			if (tmp < next_time)
				next_time = tmp;
		} else {
#ifdef NETROCKET_DEBUG_KEEPALIVE
			printk("->VS_STATE_MAYBE_DEAD...");
#endif
			vs->state = VS_STATE_MAYBE_DEAD;
			hdlc_send_ack(vs, CONTROL_IFRAME);
			vs->keep_alive_timer = jiffies + KEEP_ALIVE_TIME;
			if (KEEP_ALIVE_TIME < next_time)
				next_time = KEEP_ALIVE_TIME;
		}
	}
	/*
	 * Check the state again in case we declared the device as
	 * dead.
	 */
	if (vs->state == VS_STATE_CONNECTED) {
		/*
		 * If there are characters in the receive buffer,
		 * try to send them to the line discpline.
		 */
		for (i = 0; i < vs->num_ports; i++) 
			send_recv_buf_to_ldisc(nrp_table[i + vs->base_port]);

		/*
		 * See if we need to send commands to the remote.
		 */
		if (vs->need_to_send)
			send_remote_cmds(vs);
	}
	/*
	 * If we haven't transmited any HDLC packets within the last
	 * second, send a bare ack packet to prevent the remote from
	 * resending packets at us.
	 */
	if (vs->hdlc_state_flags & HDLC_RCV_TIMER_ACTIVE) {
#ifdef NETROCKET_DEBUG_TIMER
		printk("rcv timer...");
#endif
		tmp = vs->receiver_ack_timer - jiffies;
		if (tmp > 0) {
			if (tmp < next_time)
				next_time = tmp;
		} else {
#ifdef NETROCKET_DEBUG_TIMER
			printk("timeout...");
#endif
			hdlc_send_ack(vs, CONTROL_UFRAME);
		}
	}
	/*
	 * If we haven't gotten an ack for the packets which we have
	 * sent, resend the packets in the retransmit queue
	 */
	if (vs->hdlc_state_flags & HDLC_SND_TIMER_ACTIVE) {
#if 0
		printk("ack_timer: %ld, %ld...", vs->sender_ack_timer,
		       jiffies);
#endif
		tmp = vs->sender_ack_timer - jiffies;
		if (tmp > 0) {
			if (tmp < next_time)
				next_time = tmp;
		} else {
			hdlc_retransmit(vs);
			if (vs->hdlc_state_flags & HDLC_SND_TIMER_ACTIVE) {
				vs->sender_ack_timer = jiffies + 2*HZ;
				if (2*HZ < next_time)
					next_time = 2*HZ;
			}
		}
	}
	if (next_time != T_INFINITE)
		vs1000_set_timer(vs, next_time);

	atomic_inc(&vs->timer_active);
	
#ifdef NETROCKET_DEBUG_TIMER
	printk("End vs1000_timer (next_time = %d)\n", next_time);
#endif
}

static void vs1000_set_timer(struct vs1000_state *vs, int ticks)
{
	unsigned long new_time;
	
#ifdef NETROCKET_DEBUG_TIMER
	printk("set_timer(%d)...", ticks);
#endif
	if (!vs->timer.prev) {
		vs->timer.expires = jiffies + ticks;
		add_timer(&vs->timer);
		return;
	}
	new_time = jiffies + ticks;
	if (vs->timer.expires == new_time)
		return;
	if ((vs->timer.expires - jiffies) > ticks) {
		del_timer(&vs->timer);
		vs->timer.expires = new_time;
		add_timer(&vs->timer);
	}
}



/*****************************************************************
 * TTY interface layer
 */

/*
 * tmp_buf is used as a temporary buffer by rp_write.  We need to
 * lock it in case the memcpy_fromfs blocks while swapping in a page,
 * and some other program tries to do a serial write at the same time.
 * Since the lock will only come under contention when the system is
 * swapping and available memory is low, it makes sense to share one
 * buffer across all the serial ports, since it significantly saves
 * memory if large numbers of serial ports are open.
 */
static unsigned char *tmp_buf = 0;
static struct semaphore tmp_buf_sem = MUTEX;

/*
 * Utility function.
 */
#ifndef MIN
#define MIN(a,b)	((a) < (b) ? (a) : (b))
#endif

/*
 * This function initializes the nr_port structure
 */
static void init_nr_port(struct vs1000_state *vs, int portnum)
{
	struct nr_port *info;
	int line;
	
	line = vs->base_port + portnum;

	info = kmalloc(sizeof(struct nr_port), GFP_KERNEL);
	if (!info) {
		printk("Couldn't allocate info struct for line #%d\n", line);
		return;
	}
	memset(info, 0, sizeof(struct nr_port));
	
	info->magic = RPORT_MAGIC;
	info->line = line;
	info->vs = vs;
	info->portnum = portnum;
	info->closing_wait = 3000;
	info->close_delay = 50;
	info->baud_rate = 9600;
	info->callout_termios =callout_driver.init_termios;
	info->normal_termios = netrocket_driver.init_termios;

#if 0
	info->intmask = RXF_TRIG | TXFIFO_MT | SRC_INT | DELTA_CD |
		DELTA_CTS | DELTA_DSR;
	if (sInitChan(ctlp, &info->channel, aiop, chan) == 0) {
		printk("Rocketport sInitChan(%d, %d, %d) failed!\n",
		       board, aiop, chan);
		kfree(info);
		return;
	}
	cp = &info->channel;
#endif
	nrp_table[line] = info;
}

/*
 * If there is characters in the recv_buffer, try to send it out to
 * the line discpline
 */
static void send_recv_buf_to_ldisc(struct nr_port *info)
{
	struct tty_struct *tty;
	int	c, total_sent = 0;
	struct vs1000_port_state *port_state;

	if (!info || !info->tty || !info->recv_cnt)
		return;

	tty = info->tty;

	/* If there's nothing to send or no space, just return */
	while (1) {
		c = tty->ldisc.receive_room(tty);
		if (c > info->recv_cnt)
			c = info->recv_cnt;
		if (c > (RECV_BUF_SIZE - info->recv_tail))
			c = RECV_BUF_SIZE - info->recv_tail;
		if (c <= 0)
			break;
		tty->ldisc.receive_buf(tty, info->recv_buf + info->recv_tail, 
				       NULL, c);
		info->recv_tail += c;
		info->recv_tail &= RECV_BUF_SIZE-1;
		info->recv_cnt -= c;
		total_sent += c;
	}
	if (total_sent) {
		port_state = info->vs->port_state + info->line;
		port_state->nGetLocal += total_sent;
		port_state->update_queue = 1;
		info->vs->need_to_send = 1;
		/*
		 * We don't need to set the timer since we're called
		 * by vs1000_timer(), and the code to send the remote
		 * cmds happens immediately after the calll to this
		 * function.
		 */
	}
	/*
	 * If there are still more characters to send, try again in
	 * 0.5 seconds later.
	 */
	if (info->recv_cnt)
		vs1000_set_timer(info->vs, HZ/2);
}

/*
 * Push characters to the recv_buffer
 */
static void push_chars_to_recv_buf(struct nr_port *info,
				   unsigned char *cp, int len)
{
	struct tty_struct *tty;
	int	c;

	if (!info || !info->tty)
		return;

	tty = info->tty;

#if 0
	printk("pushed %d chars to recv_buf...", len);
#endif

	while (1) {
		c = RECV_BUF_SIZE - info->recv_cnt - 1;
		if (c > len)
			c = len;
		if (c > RECV_BUF_SIZE - info->recv_head)
			c = RECV_BUF_SIZE - info->recv_head;
		if (c <= 0)
			break;
		memcpy(info->recv_buf + info->recv_head, cp, c);
		info->recv_head = (info->recv_head + c) & (RECV_BUF_SIZE-1);
		info->recv_cnt += c;
		cp += c;
		len -= c;
	}
	info->vs->need_to_send = 1;
	vs1000_timer((unsigned long) info->vs);
}

static inline int netrocket_paranoia_check(struct nr_port *info,
					   kdev_t device, const char *routine)
{
#ifdef NETROCKET_PARANOIA_CHECK
	static const char *badmagic =
		"Warning: bad magic number for rocketport struct (%d, %d) in %s\n";
	if (!info)
		return 1;
	if (info->magic != RPORT_MAGIC) {
		printk(badmagic, MAJOR(device), MINOR(device), routine);
		return 1;
	}
#endif
	return 0;
}
	
static int baud_table[] = {
	0, 50, 75, 110, 134, 150, 200, 300,
	600, 1200, 1800, 2400, 4800, 9600, 19200,
	38400, 57600, 115200, 230400, 460800, 0 };

/*
 * This routine configures a rocketport port so according to its
 * termio settings.
 */
static void configure_nr_port(struct nr_port *info)
{
	unsigned cflag;
	int	i, bits, baud;
	int	control_settings = 0;
	
	if (!info->tty || !info->tty->termios)
		return;
	cflag = info->tty->termios->c_cflag;

	/* Byte size and parity */
	if ((cflag & CSIZE) == CS8) {
		control_settings |= SC_DATABITS_8;
		bits = 10;
	} else {
		control_settings |= SC_DATABITS_7;
		bits = 9;
	}
        if (cflag & CSTOPB) {
		control_settings |= SC_STOPBITS_2;
		bits++;
	} else {
		control_settings |= SC_STOPBITS_1;
	}
	
	if (cflag & PARENB) {
		bits++;
		if (cflag & PARODD)
			control_settings |= SC_PARITY_ODD;
		else
			control_settings |= SC_PARITY_EVEN;
	}
	if (cflag & CRTSCTS)
		control_settings |= SC_FLOW_CTS_AUTO;
	control_settings |= SC_FLOW_RTS_AUTO;

	if (control_settings != info->control_settings) {
		info->control_settings = control_settings;
		info->update_control_settings = 1;
	}
	
	/* baud rate */
	i = cflag & CBAUD;
	if (i & CBAUDEX) {
		i &= ~CBAUDEX;
		if (i < 1 || i > 4) 
			info->tty->termios->c_cflag &= ~CBAUDEX;
		else
			i += 15;
	}
	if (i == 15) {
		if ((info->flags & VS1000_SPD_MASK) == VS1000_SPD_HI)
			i += 1;
		if ((info->flags & VS1000_SPD_MASK) == VS1000_SPD_VHI)
			i += 2;
		if ((info->flags & VS1000_SPD_MASK) == VS1000_SPD_SHI)
			i += 3;
		if ((info->flags & VS1000_SPD_MASK) == VS1000_SPD_WARP)
			i += 4;
	}
	baud = baud_table[i] ? baud_table[i] : 9600;
	info->cps = baud / bits;

	if (baud != info->baud_rate) {
		info->baud_rate= baud;
		info->update_baud_rate = 1;
	}

	if (info->update_baud_rate || info->update_control_settings) {
		info->vs->need_to_send = 1;
		vs1000_timer((unsigned long) info->vs);
	}
}

static int block_til_ready(struct tty_struct *tty, struct file * filp,
			   struct nr_port *info)
{
	struct wait_queue wait = { current, NULL };
	int		retval;
	int		do_clocal = 0, extra_count = 0;
	unsigned long	flags;

	/*
	 * If the device is in the middle of being closed, then block
	 * until it's done, and then try again.
	 */
	if (tty_hung_up_p(filp))
		return ((info->flags & VS1000_HUP_NOTIFY) ? 
			-EAGAIN : -ERESTARTSYS);
	if (info->flags & VS1000_CLOSING) {
		interruptible_sleep_on(&info->close_wait);
		return ((info->flags & VS1000_HUP_NOTIFY) ? 
			-EAGAIN : -ERESTARTSYS);
	}

	/*
	 * If this is a callout device, then just make sure the normal
	 * device isn't being used.
	 */
	if (tty->driver.subtype == SERIAL_TYPE_CALLOUT) {
		if (info->flags & VS1000_NORMAL_ACTIVE)
			return -EBUSY;
		if ((info->flags & VS1000_CALLOUT_ACTIVE) &&
		    (info->flags & VS1000_SESSION_LOCKOUT) &&
		    (info->session != current->session))
		    return -EBUSY;
		if ((info->flags & VS1000_CALLOUT_ACTIVE) &&
		    (info->flags & VS1000_PGRP_LOCKOUT) &&
		    (info->pgrp != current->pgrp))
		    return -EBUSY;
		info->flags |= VS1000_CALLOUT_ACTIVE;
		return 0;
	}
	
	/*
	 * If non-blocking mode is set, or the port is not enabled,
	 * then make the check up front and then exit.
	 */
	if ((filp->f_flags & O_NONBLOCK) ||
	    (tty->flags & (1 << TTY_IO_ERROR))) {
		if (info->flags & VS1000_CALLOUT_ACTIVE)
			return -EBUSY;
		info->flags |= VS1000_NORMAL_ACTIVE;
		return 0;
	}

	if (info->flags & VS1000_CALLOUT_ACTIVE) {
		if (info->normal_termios.c_cflag & CLOCAL)
			do_clocal = 1;
	} else {
		if (tty->termios->c_cflag & CLOCAL)
			do_clocal = 1;
	}
	
	/*
	 * Block waiting for the carrier detect and the line to become
	 * free (i.e., not in use by the callout).  While we are in
	 * this loop, info->count is dropped by one, so that
	 * nrp_close() knows when to free things.  We restore it upon
	 * exit, either normal or abnormal.
	 */
	retval = 0;
	add_wait_queue(&info->open_wait, &wait);
#ifdef NETROCKET_DEBUG_OPEN
	printk("block_til_ready before block: ttyR%d, count = %d\n",
	       info->line, info->count);
#endif
	save_flags(flags); cli();
	if (!tty_hung_up_p(filp)) {
		extra_count = 1;
		info->count--;
	}
	restore_flags(flags);
	info->blocked_open++;
	while (1) {
		if (!(info->flags & VS1000_CALLOUT_ACTIVE) &&
		    (tty->termios->c_cflag & CBAUD)) {
			if (!(info->modem_control_register &
			      MCR_DTR_SET_MASK)) {
				info->modem_control_register |= MCR_DTR_SET_ON|
					MCR_RTS_SET_ON;
				info->update_mcr = 1;
				info->vs->need_to_send = 1;
				vs1000_timer((unsigned long) info->vs);
			}
		}
		current->state = TASK_INTERRUPTIBLE;
		if (tty_hung_up_p(filp) ||
		    !(info->flags & VS1000_INITIALIZED)) {
			if (info->flags & VS1000_HUP_NOTIFY)
				retval = -EAGAIN;
			else
				retval = -ERESTARTSYS;	
			break;
		}
#if 0
		if (!(info->flags & VS1000_CALLOUT_ACTIVE) &&
		    !(info->flags & VS1000_CLOSING) &&
		    (do_clocal || (sGetChanStatusLo(&info->channel) &
				   CD_ACT)))
			break;
#else
		break;
#endif
		if (signal_pending(current)) {
			retval = -ERESTARTSYS;
			break;
		}
#ifdef NETROCKET_DEBUG_OPEN
		printk("block_til_ready blocking: ttyR%d, count = %d, flags=0x%0x\n",
		       info->line, info->count, info->flags);
#endif
		schedule();
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&info->open_wait, &wait);
	cli();
	if (extra_count)
		info->count++;
	restore_flags(flags);
	info->blocked_open--;
#ifdef NETROCKET_DEBUG_OPEN
	printk("block_til_ready after blocking: ttyR%d, count = %d\n",
	       info->line, info->count);
#endif
	if (retval)
		return retval;
	info->flags |= VS1000_NORMAL_ACTIVE;
	return 0;
}	

/*
 * This routine is called whenever a rocketport board is opened.
 */
static int nrp_open(struct tty_struct *tty, struct file * filp)
{
	struct vs1000_port_state *pstate;
	struct nr_port *info;
	int	line, retval;
	unsigned long page, page2;

	line = MINOR(tty->device) - tty->driver.minor_start;
	if ((line < 0) || (line >= MAX_NRP_PORTS))
		return -ENODEV;
	if (!tmp_buf) {
		page = get_free_page(GFP_KERNEL);
		if (!page)
			return -ENOMEM;
		if (tmp_buf)
			free_page(page);
		else
			tmp_buf = (unsigned char *) page;
	}
	page = get_free_page(GFP_KERNEL);
	if (!page)
		return -ENOMEM;

	page2 = get_free_page(GFP_KERNEL);
	if (!page) {
		free_page(page);
		return -ENOMEM;
	}

	tty->driver_data = info = nrp_table[line];
	
	if (info == NULL) {
		tty->flags = (1 << TTY_IO_ERROR);
		free_page(page2);
		free_page(page);
		return 0;
	}
	
	if (info->flags & VS1000_CLOSING) {
		interruptible_sleep_on(&info->close_wait);
		free_page(page2);
		free_page(page);
		return ((info->flags & VS1000_HUP_NOTIFY) ?
			-EAGAIN : -ERESTARTSYS);
	}
	/*
	 * We must not sleep from here until the port is marked fully
	 * in use.
	 */
	if (info->xmit_buf)
		free_page(page);
	else
		info->xmit_buf = (unsigned char *) page;
	if (info->recv_buf)
		free_page(page2);
	else
		info->recv_buf = (unsigned char *) page2;
	info->tty = tty;

	if (info->flags & VS1000_CLOSING) {
		interruptible_sleep_on(&info->close_wait);
		return ((info->flags & VS1000_HUP_NOTIFY) ? 
			-EAGAIN : -ERESTARTSYS);
	}

	if (info->count++ == 0) {
#ifdef MODULE
		MOD_INC_USE_COUNT;
#endif
		nrp_num_ports_open++;
#ifdef NETROCKET_DEBUG_OPEN
		printk("rocket mod++ = %d...", nrp_num_ports_open);
#endif
	}
#ifdef NETROCKET_DEBUG_OPEN
	printk("nrp_open ttyR%d, count=%d\n", info->line, info->count);
#endif
	/*
	 * Info->count is now 1; so it's safe to sleep now.
	 */
	info->session = current->session;
	info->pgrp = current->pgrp;

	pstate = info->vs->port_state + info->portnum;
	if (pstate->modemStatusRegister & MSR_CD_ON)
		info->cd_status = 1;
	else
		info->cd_status = 0;

	/*
	 * XXX Shouldn't be needed, but embedded box sometimes drops a
	 * transmitted data subpacket request (while still ack'ing the
	 * packet at the HDLC level!)  We resync on an initial open to
	 * make sure things are sane...
	 */
	pstate->nPutRemote = pstate->nGetRemote;

	info->action_register = ACT_FLUSH_INPUT | ACT_FLUSH_OUTPUT;
	info->update_action = 1;
	info->vs->need_to_send = 1;
	
#if (LINUX_VERSION_CODE >= 131394) /* Linux 2.1.66 */
	/*
	 * Set up the tty->alt_speed kludge
	 */
	if ((info->flags & VS1000_SPD_MASK) == VS1000_SPD_HI)
		info->tty->alt_speed = 57600;
	if ((info->flags & VS1000_SPD_MASK) == VS1000_SPD_VHI)
		info->tty->alt_speed = 115200;
	if ((info->flags & VS1000_SPD_MASK) == VS1000_SPD_SHI)
		info->tty->alt_speed = 230400;
	if ((info->flags & VS1000_SPD_MASK) == VS1000_SPD_WARP)
		info->tty->alt_speed = 460800;
#endif

	configure_nr_port(info);
	if (tty->termios->c_cflag & CBAUD)
		info->modem_control_register |= MCR_DTR_SET_ON;
	info->modem_control_register |= MCR_RTS_SET_ON;
	info->update_mcr = 1;
	info->vs->need_to_send = 1;
	vs1000_timer((unsigned long) info->vs);

	info->flags |= VS1000_INITIALIZED;

	retval = block_til_ready(tty, filp, info);
	if (retval) {
#ifdef NETROCKET_DEBUG_OPEN
		printk("nrp_open returning after block_til_ready with %d\n",
		       retval);
#endif
		return retval;
	}

	if ((info->count == 1) && (info->flags & VS1000_SPLIT_TERMIOS)) {
		if (tty->driver.subtype == SERIAL_TYPE_NORMAL)
			*tty->termios = info->normal_termios;
		else 
			*tty->termios = info->callout_termios;
		configure_nr_port(info);
	}

	return 0;
}

static void nrp_close(struct tty_struct *tty, struct file * filp)
{
	struct vs1000_port_state *pstate;
	struct nr_port * info = (struct nr_port *)tty->driver_data;
	unsigned long flags;
	int timeout;

	if (netrocket_paranoia_check(info, tty->device, "nrp_close"))
		return;

	if (!info)
		return;

#ifdef NETROCKET_DEBUG_OPEN
	printk("nrp_close ttyR%d, count = %d\n", info->line, info->count);
#endif
	
	save_flags(flags); cli();
	
	if (tty_hung_up_p(filp)) {
		restore_flags(flags);
		return;
	}
	if ((tty->count == 1) && (info->count != 1)) {
		/*
		 * Uh, oh.  tty->count is 1, which means that the tty
		 * structure will be freed.  Info->count should always
		 * be one in these conditions.  If it's greater than
		 * one, we've got real problems, since it means the
		 * serial port won't be shutdown.
		 */
		printk("nrp_close: bad serial port count; tty->count is 1, "
		       "info->count is %d\n", info->count);
		info->count = 1;
	}
	if (--info->count < 0) {
		printk("nrp_close: bad serial port count for ttyR%d: %d\n",
		       info->line, info->count);
		info->count = 0;
	}
	if (info->count) {
		restore_flags(flags);
		return;
	}
	info->flags |= VS1000_CLOSING;
	/*
	 * Save the termios structure, since this port may have
	 * separate termios for callout and dialin.
	 */
	if (info->flags & VS1000_NORMAL_ACTIVE)
		info->normal_termios = *tty->termios;
	if (info->flags & VS1000_CALLOUT_ACTIVE)
		info->callout_termios = *tty->termios;
	
	/*
	 * Notify the line discpline to only process XON/XOFF characters
	 */
	tty->closing = 1;

	/*
	 * If transmission was throttled by the application request,
	 * just flush the xmit buffer.
	 */
#if (LINUX_VERSION_CODE >= 131343)
	if (tty->flow_stopped)
		nrp_flush_buffer(tty);
#endif

	/*
	 * Wait for the transmit buffer to clear
	 */
	if (info->closing_wait != VS1000_CLOSING_WAIT_NONE)
		tty_wait_until_sent(tty, info->closing_wait);
	/*
	 * Before we drop DTR, make sure the UART transmitter
	 * has completely drained; this is especially
	 * important if there is a transmit FIFO!
	 */
	pstate = info->vs->port_state + info->portnum;
	timeout = (get_tx_cnt_remote(pstate) + 1) * HZ / info->cps;
	if (timeout == 0)
		timeout = 1;
	nrp_wait_until_sent(tty, timeout);

	info->action_register = ACT_FLUSH_INPUT | ACT_FLUSH_OUTPUT;
	info->update_action = 1;

	info->modem_control_register &= ~MCR_RTS_SET_MASK;
	info->update_mcr = 1;
	
	if (C_HUPCL(tty)) {
		if (info->modem_control_register & MCR_DTR_SET_MASK) {
			info->modem_control_register &= ~MCR_DTR_SET_MASK;
			info->update_mcr = 1;
		}
	}

	info->vs->need_to_send = 1;
	vs1000_timer((unsigned long) info->vs);
	
	if (tty->driver.flush_buffer)
		tty->driver.flush_buffer(tty);
	if (tty->ldisc.flush_buffer)
		tty->ldisc.flush_buffer(tty);

	if (info->blocked_open) {
		if (info->close_delay) {
			current->state = TASK_INTERRUPTIBLE;
			current->timeout = jiffies + info->close_delay;
			schedule();
		}
		wake_up_interruptible(&info->open_wait);
	} else {
		if (info->xmit_buf) {
			free_page((unsigned long) info->xmit_buf);
			info->xmit_buf = 0;
		}
		if (info->recv_buf) {
			free_page((unsigned long) info->recv_buf);
			info->recv_buf = 0;
		}
	}
	info->flags &= ~(VS1000_INITIALIZED | VS1000_CLOSING |
			 VS1000_CALLOUT_ACTIVE | VS1000_NORMAL_ACTIVE);
	tty->closing = 0;
	wake_up_interruptible(&info->close_wait);
	
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
	nrp_num_ports_open--;
#ifdef NETROCKET_DEBUG_OPEN
	printk("rocket mod-- = %d...", nrp_num_ports_open);
#endif
	restore_flags(flags);
	
#ifdef NETROCKET_DEBUG_OPEN
	printk("nrp_close ttyR%d complete shutdown\n", info->line);
#endif
	
}

static void nrp_set_termios(struct tty_struct *tty, struct termios *old_termios)
{
	struct nr_port * info = (struct nr_port *)tty->driver_data;
	unsigned cflag;
	

	if (netrocket_paranoia_check(info, tty->device, "nrp_set_termios"))
		return;

	cflag = tty->termios->c_cflag;

	if (cflag == old_termios->c_cflag)
		return;

	/*
	 * This driver doesn't support CS5 or CS6
	 */
	if (((cflag & CSIZE) == CS5) ||
	    ((cflag & CSIZE) == CS6))
		tty->termios->c_cflag = ((cflag & ~CSIZE) |
					 (old_termios->c_cflag & CSIZE));

	configure_nr_port(info);

	/* Handle transition to B0 status */
	if ((old_termios->c_cflag & CBAUD) &&
	    !(tty->termios->c_cflag & CBAUD)) {
		info->modem_control_register &=
			~(MCR_RTS_SET_MASK|MCR_DTR_SET_MASK);
		info->update_mcr = 1;
	}
	
	/* Handle transition away from B0 status */
	if (!(old_termios->c_cflag & CBAUD) &&
	    (tty->termios->c_cflag & CBAUD)) {
		info->modem_control_register |=
			MCR_RTS_SET_ON | MCR_DTR_SET_ON;
		info->update_mcr = 1;
	}
	
	if ((old_termios->c_cflag & CRTSCTS) &&
	    !(tty->termios->c_cflag & CRTSCTS)) {
		tty->hw_stopped = 0;
		nrp_start(tty);
	}
}

/*
 * Here are the routines used by nrp_ioctl
 */
static void nrp_break(struct tty_struct *tty, int break_state)
{
	struct nr_port * info = (struct nr_port *)tty->driver_data;
	int		orig;
	
	if (netrocket_paranoia_check(info, tty->device, "nrp_break"))
		return;

	orig = info->modem_control_register;
	if (break_state == -1)
		info->modem_control_register |= MCR_BREAK_SET_ON;
	else
		info->modem_control_register &= ~MCR_BREAK_SET_ON;
	if (orig != info->modem_control_register) {
		info->update_mcr = 1;
	}
}

#if (LINUX_VERSION_CODE < 131394) /* Linux 2.1.66 */
static void send_break(	struct nr_port * info, int duration)
{
	current->state = TASK_INTERRUPTIBLE;
	current->timeout = jiffies + duration;
	nrp_break(info->tty, -1);
	schedule();
	nrp_break(info->tty, 0);
}
#endif


static int get_modem_info(struct nr_port * info, unsigned int *value)
{
	unsigned int result;
	int	mcr, msr;

	mcr = info->modem_control_register;
	msr = info->vs->port_state[info->line].modemStatusRegister;

	result =  ((mcr & MCR_RTS_SET_ON) ? TIOCM_RTS : 0)
		| ((mcr & MCR_DTR_SET_ON) ? TIOCM_DTR : 0)
		| ((msr & MSR_CD_ON) ? TIOCM_CAR : 0)
			/* TIOCM_RNG not supported */
		| ((msr & MSR_DSR_ON) ? TIOCM_DSR : 0)
		| ((msr & MSR_CTS_ON) ? TIOCM_CTS : 0);

	if (copy_to_user(value, &result, sizeof(int)))
		return -EFAULT;
	return 0;
}

static int set_modem_info(struct nr_port * info, unsigned int cmd,
			  unsigned int *value)
{
	unsigned int arg;
	int	mcr;

	if (copy_from_user(&arg, value, sizeof(int)))
		return -EFAULT;

	mcr = info->modem_control_register;

	switch (cmd) {
	case TIOCMBIS: 
		if (arg & TIOCM_RTS)
			mcr |= MCR_RTS_SET_ON;
		if (arg & TIOCM_DTR)
			mcr |= MCR_DTR_SET_ON;
		break;
	case TIOCMBIC:
		if (arg & TIOCM_RTS)
			mcr &= ~MCR_RTS_SET_ON;
		if (arg & TIOCM_DTR)
			mcr &= ~MCR_DTR_SET_ON;
		break;
	case TIOCMSET:
		mcr = ((mcr & ~(MCR_RTS_SET_ON | MCR_DTR_SET_ON))
		       | ((arg & TIOCM_RTS) ? MCR_RTS_SET_ON : 0)
		       | ((arg & TIOCM_DTR) ? MCR_DTR_SET_ON : 0));
		break;
	default:
		return -EINVAL;
	}

	if (mcr != info->modem_control_register) {
		info->modem_control_register = mcr;
		info->update_mcr = 1;
		info->vs->need_to_send = 1;
		vs1000_timer((unsigned long) info->vs);
	}

	return 0;
}

static int get_config(struct nr_port * info, struct vs1000_config * retinfo)
{
	struct vs1000_config tmp;
  
	if (!retinfo)
		return -EFAULT;
	memset(&tmp, 0, sizeof(tmp));
	tmp.line = info->line;
	tmp.flags = info->flags;
	tmp.close_delay = info->close_delay;
	tmp.closing_wait = info->closing_wait;
#if 0
	tmp.port = rcktpt_io_addr[(info->line >> 5) & 3];
#else
	tmp.port = 0;
#endif
	
	if (copy_to_user(retinfo,&tmp,sizeof(*retinfo)))
		return -EFAULT;
	return 0;
}

static int set_config(struct nr_port * info, struct vs1000_config * new_info)
{
	struct vs1000_config new_serial;

	if (copy_from_user(&new_serial, new_info, sizeof(new_serial)))
		return -EFAULT;

	if (!suser()) {
		if ((new_serial.flags & ~VS1000_USR_MASK) !=
		    (info->flags & ~VS1000_USR_MASK))
			return -EPERM;
		info->flags = ((info->flags & ~VS1000_USR_MASK) |
			       (new_serial.flags & VS1000_USR_MASK));
		configure_nr_port(info);
		return 0;
	}
	
	info->flags = ((info->flags & ~VS1000_FLAGS) |
			(new_serial.flags & VS1000_FLAGS));
	info->close_delay = new_serial.close_delay;
	info->closing_wait = new_serial.closing_wait;

#if (LINUX_VERSION_CODE >= 131394) /* Linux 2.1.66 */
	if ((info->flags & VS1000_SPD_MASK) == VS1000_SPD_HI)
		info->tty->alt_speed = 57600;
	if ((info->flags & VS1000_SPD_MASK) == VS1000_SPD_VHI)
		info->tty->alt_speed = 115200;
	if ((info->flags & VS1000_SPD_MASK) == VS1000_SPD_SHI)
		info->tty->alt_speed = 230400;
	if ((info->flags & VS1000_SPD_MASK) == VS1000_SPD_WARP)
		info->tty->alt_speed = 460800;
#endif
	
	configure_nr_port(info);
	return 0;
}

static int nrp_ioctl(struct tty_struct *tty, struct file * file,
		    unsigned int cmd, unsigned long arg)
{
	struct nr_port * info = (struct nr_port *)tty->driver_data;
#if (LINUX_VERSION_CODE < 131394) /* Linux 2.1.66 */
	int retval, tmp;
#endif

	if (netrocket_paranoia_check(info, tty->device, "nrp_ioctl"))
		return -ENODEV;

	switch (cmd) {
#if (LINUX_VERSION_CODE < 131394) /* Linux 2.1.66 */
		case TCSBRK:	/* SVID version: non-zero arg --> no break */
			retval = tty_check_change(tty);
			if (retval)
				return retval;
			tty_wait_until_sent(tty, 0);
			if (signal_pending(current))
				return -EINTR;
			if (!arg) {
				send_break(info, HZ/4);	/* 1/4 second */
				if (signal_pending(current))
					return -EINTR;
			}
			return 0;
		case TCSBRKP:	/* support for POSIX tcsendbreak() */
			retval = tty_check_change(tty);
			if (retval)
				return retval;
			tty_wait_until_sent(tty, 0);
			if (signal_pending(current))
				return -EINTR;
			send_break(info, arg ? arg*(HZ/10) : HZ/4);
			if (signal_pending(current))
				return -EINTR;
			return 0;
		case TIOCGSOFTCAR:
			tmp = C_CLOCAL(tty) ? 1 : 0;
			if (copy_to_user((void *)arg, &tmp, sizeof(int)))
				return -EFAULT;
			return 0;
		case TIOCSSOFTCAR:
			if (copy_from_user(&tmp, (void *)arg, sizeof(int)))
				return -EFAULT;

			tty->termios->c_cflag =
				((tty->termios->c_cflag & ~CLOCAL) |
				 (tmp ? CLOCAL : 0));
			return 0;
#endif
		case TIOCMGET:
			return get_modem_info(info, (unsigned int *) arg);
		case TIOCMBIS:
		case TIOCMBIC:
		case TIOCMSET:
			return set_modem_info(info, cmd, (unsigned int *) arg);
		case VS1000_LOAD_CODE:
			return send_boot_packet(info->vs, (void *) arg);
		case VS1000_GET_CONFIG:
			return get_config(info, (struct vs1000_config *) arg);
		case VS1000_SET_CONFIG:
			return set_config(info, (struct vs1000_config *) arg);
		case VS1000_GET_CTL_MAJOR:
			if (copy_to_user((void *)arg, &control_major,
					 sizeof(int)))
				return -EFAULT;
		case VS1000_SET_BOX_CONFIG:
		case VS1000_GET_BOX_CONFIG:
		case VS1000_GET_BOX_REPORT:
			return vs_ctl_ioctl(NULL, file, cmd, arg);
			
		default:
			return -ENOIOCTLCMD;
		}
	return 0;
}

#if (defined(NETROCKET_DEBUG_FLOW) || defined(NETROCKET_DEBUG_THROTTLE))
static char *nrp_tty_name(struct tty_struct *tty, char *buf)
{
	if (tty)
		sprintf(buf, "%s%d", tty->driver.name,
			MINOR(tty->device) - tty->driver.minor_start +
			tty->driver.name_base);
	else
		strcpy(buf, "NULL tty");
	return buf;
}
#endif

static void nrp_send_xchar(struct tty_struct *tty, char ch)
{
	struct nr_port *info = (struct nr_port *)tty->driver_data;

	if (netrocket_paranoia_check(info, tty->device, "nrp_send_xchar"))
		return;

	info->x_char = ch;
	info->vs->need_to_send = 1;
	vs1000_timer((unsigned long) info->vs);
}

static void nrp_throttle(struct tty_struct * tty)
{
	struct nr_port *info = (struct nr_port *)tty->driver_data;
#ifdef NETROCKET_DEBUG_THROTTLE
	char	buf[64];
	
	printk("throttle %s: %d....\n", nrp_tty_name(tty, buf),
	       tty->ldisc.chars_in_buffer(tty));
#endif

	if (netrocket_paranoia_check(info, tty->device, "nrp_throttle"))
		return;

	if (I_IXOFF(tty))
		nrp_send_xchar(tty, STOP_CHAR(tty));
	
#if 0
	sClrRTS(&info->channel);
#endif
}

static void nrp_unthrottle(struct tty_struct * tty)
{
	struct nr_port *info = (struct nr_port *)tty->driver_data;
#ifdef NETROCKET_DEBUG_THROTTLE
	char	buf[64];
	
	printk("unthrottle %s: %d....\n", nrp_tty_name(tty, buf),
	       tty->ldisc.chars_in_buffer(tty));
#endif

	if (netrocket_paranoia_check(info, tty->device, "nrp_throttle"))
		return;

	if (I_IXOFF(tty))
		nrp_send_xchar(tty, START_CHAR(tty));

#if 0
	sSetRTS(&info->channel);
#endif
}

/*
 * ------------------------------------------------------------
 * nrp_stop() and nrp_start()
 *
 * This routines are called before setting or resetting tty->stopped.
 * They enable or disable transmitter interrupts, as necessary.
 * ------------------------------------------------------------
 */
static void nrp_stop(struct tty_struct *tty)
{
	struct nr_port * info = (struct nr_port *)tty->driver_data;
#ifdef NETROCKET_DEBUG_FLOW
	char	buf[64];
	
	printk("stop %s: %d %d....\n", nrp_tty_name(tty, buf),
	       info->xmit_cnt, info->xmit_fifo_room);
#endif

	if (netrocket_paranoia_check(info, tty->device, "nrp_stop"))
		return;

	/*
	 * Would be nice if there was a way to stop the transmitter
	 * immediately.
	 */
}

static void nrp_start(struct tty_struct *tty)
{
	struct nr_port * info = (struct nr_port *)tty->driver_data;
#ifdef NETROCKET_DEBUG_FLOW
	char	buf[64];
	
	printk("start %s: %d %d....\n", nrp_tty_name(tty, buf),
	       info->xmit_cnt, info->xmit_fifo_room);
#endif

	if (netrocket_paranoia_check(info, tty->device, "nrp_start"))
		return;

	/*
	 * Start the transmitter remotely, if we ever add a stop
	 * transmitter to nrp_stop.
	 */

	if (info->xmit_cnt <= 0 || tty->stopped || tty->hw_stopped ||
	    !info->xmit_buf)
		return;
	
	info->vs->need_to_send = 1;
	vs1000_timer((unsigned long) info->vs);
}

/*
 * nrp_wait_until_sent() --- wait until the transmitter is empty
 */
static void nrp_wait_until_sent(struct tty_struct *tty, int timeout)
{
	struct nr_port *info = (struct nr_port *)tty->driver_data;
	unsigned long orig_jiffies;
	int check_time, exit_time;
	int txcnt;
	
	if (netrocket_paranoia_check(info, tty->device, "nrp_wait_until_sent"))
		return;

	orig_jiffies = jiffies;
#ifdef NETROCKET_DEBUG_WAIT_UNTIL_SENT
	printk("In NRP_wait_until_sent(%d) (jiff=%lu)...", timeout, jiffies);
	printk("cps=%d...", info->cps);
#endif
	while (1) {
		txcnt = get_tx_cnt_remote(&info->vs->port_state[info->line]);
		if (!txcnt) {
			/*
			 * Ideal to check if the last character has
			 * shifted all the way out, but there's no way
			 * to do that right now.
			 */
			break;
		} else
			check_time = HZ * txcnt / info->cps;
		if (timeout) {
			exit_time = orig_jiffies + timeout - jiffies;
			if (exit_time <= 0)
				break;
			if (exit_time < check_time)
				check_time = exit_time;
		}
		if (check_time == 0)
			check_time = 1;
#ifdef NETROCKET_DEBUG_WAIT_UNTIL_SENT
		printk("txcnt = %d (jiff=%lu,check=%d)...", txcnt,
		       jiffies, check_time);
#endif
		current->state = TASK_INTERRUPTIBLE;
		current->counter = 0;	/* make us low-priority */
		current->timeout = jiffies + check_time;
		schedule();
		if (signal_pending(current))
			break;
	}
	current->state = TASK_RUNNING;
#ifdef NETROCKET_DEBUG_WAIT_UNTIL_SENT
	printk("txcnt = %d (jiff=%lu)...done\n", txcnt, jiffies);
#endif
}

/*
 * nrp_hangup() --- called by tty_hangup() when a hangup is signaled.
 */
static void nrp_hangup(struct tty_struct *tty)
{
	struct nr_port * info = (struct nr_port *)tty->driver_data;
	
	if (netrocket_paranoia_check(info, tty->device, "nrp_hangup"))
		return;

#if (defined(NETROCKET_DEBUG_OPEN) || defined(NETROCKET_DEBUG_HANGUP))
	printk("nrp_hangup of ttyR%d...", info->line);
#endif
	/*
	 * If the port is in the process of being closed, just force
	 * the transmit buffer to be empty, and let nrp_close handle
	 * the clean up.
	 */
	if (info->flags & VS1000_CLOSING) {
		cli();
		info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
		sti();
		wake_up_interruptible(&tty->write_wait);
		return;
	}
	if (info->count) {
#ifdef MODULE
		MOD_DEC_USE_COUNT;
#endif
		nrp_num_ports_open--;
	}
	
	info->count = 0;
	info->flags &= ~(VS1000_NORMAL_ACTIVE|VS1000_CALLOUT_ACTIVE);
	info->tty = 0;

	info->action_register = ACT_FLUSH_INPUT | ACT_FLUSH_OUTPUT;
	info->update_action = 1;
	info->vs->need_to_send = 1;

	vs1000_timer((unsigned long) info->vs);

	info->flags &= ~VS1000_INITIALIZED;
	
	wake_up_interruptible(&info->open_wait);
}

/*
 * The output routines for the vs1000.
 */
static void nrp_put_char(struct tty_struct *tty, unsigned char ch)
{
	struct nr_port * info = (struct nr_port *)tty->driver_data;
	unsigned long flags;

	if (netrocket_paranoia_check(info, tty->device, "nrp_put_char"))
		return;

#ifdef NETROCKET_DEBUG_WRITE
	printk("nrp_put_char %c...", ch);
#endif

	save_flags(flags); cli();
	if (info->xmit_cnt >= XMIT_BUF_SIZE - 1) {
		restore_flags(flags);
		return;
	}
	
	info->xmit_buf[info->xmit_head++] = ch;
	info->xmit_head &= SERIAL_XMIT_SIZE-1;
	info->xmit_cnt++;
	restore_flags(flags);
}

static void nrp_flush_chars(struct tty_struct *tty)
{
	struct nr_port * info = (struct nr_port *)tty->driver_data;
	struct vs1000_port_state *pstate;
	int	space = 0;
				
	if (netrocket_paranoia_check(info, tty->device, "nrp_flush_chars"))
		return;

	pstate = &info->vs->port_state[info->line];
	if (pstate)
		space = get_tx_space_remote(pstate);

	if (space <= WAKEUP_CHARS || info->xmit_cnt <= 0 ||
	    tty->stopped || tty->hw_stopped || !info->xmit_buf)
		return;

	info->vs->need_to_send = 1;
	if (info->xmit_cnt > 64)
		vs1000_timer((unsigned long) info->vs);
	else
		vs1000_set_timer(info->vs, 0);
}

#if (LINUX_VERSION_CODE > 66304)
static int nrp_write(struct tty_struct * tty, int from_user,
		    const unsigned char *buf, int count)
#else
static int nrp_write(struct tty_struct * tty, int from_user,
		    unsigned char *buf, int count)
#endif	
{
	struct nr_port * info = (struct nr_port *)tty->driver_data;
	const unsigned char	*b;
	int		c, retval = 0;
	unsigned long	flags;

	if (count <= 0 ||
	    netrocket_paranoia_check(info, tty->device, "nrp_write"))
		return 0;

#ifdef NETROCKET_DEBUG_WRITE
	printk("nrp_write %d chars...", count);
#endif

	save_flags(flags);
	while (1) {
		cli();
		if (info->tty == 0) {
			restore_flags(flags);
			goto end;
		}
		c = MIN(count, MIN(XMIT_BUF_SIZE - info->xmit_cnt - 1,
				   XMIT_BUF_SIZE - info->xmit_head));
		if (c <= 0)
			break;

		b = buf;
		if (from_user) {
			down(&tmp_buf_sem);
			c -= copy_from_user(tmp_buf, buf, c);
			b = tmp_buf;
			up(&tmp_buf_sem);
			if (!c) {
				if (retval == 0)
					retval = -EFAULT;
				goto end_intr;
			}
			/* In case we got pre-empted */
			if (info->tty == 0)
				goto end_intr;
			c = MIN(c, MIN(XMIT_BUF_SIZE - info->xmit_cnt - 1,
				       XMIT_BUF_SIZE - info->xmit_head));
				
		}
		memcpy(info->xmit_buf + info->xmit_head, b, c);
		info->xmit_head = (info->xmit_head + c) & (XMIT_BUF_SIZE-1);
		info->xmit_cnt += c;
		restore_flags(flags);
		buf += c;
		count -= c;
		retval += c;
	}
end_intr:
	restore_flags(flags);
	nrp_flush_chars(tty);
end:
	return retval;
}

/*
 * Return the number of characters that can be sent.  We estimate
 * only using the in-memory transmit buffer only, and ignore the
 * potential space in the remote end.
 */
static int nrp_write_room(struct tty_struct *tty)
{
	struct nr_port * info = (struct nr_port *)tty->driver_data;
	int	ret;

	if (netrocket_paranoia_check(info, tty->device, "nrp_write_room"))
		return 0;

	ret = XMIT_BUF_SIZE - info->xmit_cnt - 1;
	if (ret < 0)
		ret = 0;
#ifdef NETROCKET_DEBUG_WRITE
	printk("nrp_write_room returns %d...", ret);
#endif
	return ret;
}

/*
 * Return the number of characters in the buffer.  Again, this only
 * counts those characters in the in-memory transmit buffer.
 */
static int nrp_chars_in_buffer(struct tty_struct *tty)
{
	struct nr_port * info = (struct nr_port *)tty->driver_data;

	if (netrocket_paranoia_check(info, tty->device, "nrp_chars_in_buffer"))
		return 0;

#ifdef NETROCKET_DEBUG_WRITE
	printk("nrp_chars_in_buffer returns %d...", info->xmit_cnt);
#endif
	return info->xmit_cnt;
}

static void nrp_flush_buffer(struct tty_struct *tty)
{
	struct nr_port * info = (struct nr_port *)tty->driver_data;
	unsigned long flags;

	if (netrocket_paranoia_check(info, tty->device, "nrp_flush_buffer"))
		return;

	info->action_register = ACT_FLUSH_OUTPUT;
	info->update_action = 1;
	info->vs->need_to_send = 1;
	vs1000_timer((unsigned long) info->vs);
	
	save_flags(flags); cli();
	info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
	restore_flags(flags);
	wake_tty_write(tty);
}

/*
 * ---------------------------------------------------------------
 *  This is where we put the functions used by the control device.
 */

static int vs_ctl_open(struct inode * inode, struct file * file)
{
	MOD_INC_USE_COUNT;
	return 0;
}

#if (LINUX_VERSION_CODE >= 131343) /* 2.1.15 -- XX get correct version */
static int vs_ctl_close(struct inode * inode, struct file * file)
#else
static void vs_ctl_close(struct inode * inode, struct file * file)
#endif
{
	MOD_DEC_USE_COUNT;
#if (LINUX_VERSION_CODE >= 131343) /* 2.1.15 -- XX get correct version */
	return 0;
#else
	return;
#endif
}

#if (LINUX_VERSION_CODE >= 131343) /* 2.1.15 -- XX get correct version */
static unsigned int
vs_ctl_poll(struct file *file, poll_table * wait)
{
	struct vs1000_state 	*vs;
	unsigned int 		mask;
	int			i;
	
	poll_wait(file, &wake_daemon_wait, wait);
	mask = 0;

	for (i=0; i < NUM_VS1000; i++) {
		vs = conc_state[i];
		if (vs && vs->state == VS_STATE_DEAD)
			mask |= POLLOUT | POLLWRNORM;
	}
	return mask;
}
#else
static int
vs_ctl_select(struct inode *inode, struct file *file,
		      int sel_type, select_table * wait)
{
	struct vs1000_state 	*vs;
	int			i;
	
	switch (sel_type) {
	case SEL_OUT:
		for (i=0; i < NUM_VS1000; i++) {
			vs = conc_state[i];
			if (vs && vs->state == VS_STATE_DEAD)
				return 1;
		}
		select_wait(&wake_daemon_wait, wait);
		break;
	}
	return 0;
}
#endif

static int get_box_config(struct vs1000_box_config *retinfo)
{
	struct vs1000_box_config tmp;
	struct vs1000_state *vs;
  
	if (copy_from_user(&tmp, retinfo, sizeof(tmp)))
		return -EFAULT;

	if ((tmp.boxid >= NUM_VS1000) || !((vs = conc_state[tmp.boxid])))
		return -EINVAL;

	tmp.num_ports = vs->num_ports;
	tmp.state = vs->state;
	memcpy(tmp.ether_addr, vs->ether_addr, 6);
	strncpy(tmp.interface, vs->dev->name, sizeof(tmp.interface));
	memset(tmp.reserved, 0, sizeof(tmp.reserved));

	if (copy_to_user(retinfo,&tmp,sizeof(tmp)))
		return -EFAULT;
	return 0;
}

static int get_box_report(struct vs1000_box_report *retinfo)
{
	struct vs1000_box_report tmp;
	int	unit;
  
	if (copy_from_user(&tmp, retinfo, sizeof(tmp)))
		return -EFAULT;

	unit = tmp.flags;	/* kludge, but easist way to handle it */

	if ((unit >= num_box_rpts) || (unit >= MAX_BOX_REPORTS))
		return -EINVAL;

	tmp = vs1000_box_rpts[unit];
	
	if (copy_to_user(retinfo,&tmp,sizeof(tmp)))
		return -EFAULT;
	return 0;
}

static int get_problem_boxes(int *ret_status)
{
	struct vs1000_state *vs;
	int	i, status = 0;
	
	for (i=0; i < NUM_VS1000; i++) {
		vs = conc_state[i];
		if (vs && vs->state == VS_STATE_DEAD)
			status |= 1 << i;
	}
	if (copy_to_user(ret_status, &status, sizeof(status)))
		return -EFAULT;
	return 0;
}

static int debug_misc()
{
#if 0
	struct vs1000_state *vs;
	
	vs = conc_state[0];
	printk("vs->rtq.qlen = %d\n", vs->rtq.qlen);
#endif
	return 0;
}	

static int
vs_ctl_ioctl(struct inode * inode, struct file * file,
	     unsigned int cmd, unsigned long arg)
{
	switch (cmd) {
	case VS1000_LOAD_CODE:
		return send_boot_packet(NULL, (void *) arg);

	case VS1000_SET_BOX_CONFIG:
		return init_vs1000_state((struct vs1000_box_config *) arg, 1);
		
	case VS1000_GET_BOX_CONFIG:
		return get_box_config((struct vs1000_box_config *) arg);

	case VS1000_GET_BOX_REPORT:
		return get_box_report((struct vs1000_box_report *) arg);

	case VS1000_GET_PROBLEM_BOXES:
		return get_problem_boxes((int *) arg);

	case VS1000_DEBUG_MISC:
		return debug_misc();

	default:
		return -EINVAL;
	}
	return 0;
}

static struct file_operations ctl_fops = {
	NULL,			/* lseek */
	NULL,			/* read */
	NULL,			/* write */
	NULL,			/* readdir */
#if (LINUX_VERSION_CODE >= 131343) /* 2.1.15 -- XX get correct version */
	vs_ctl_poll,		/* poll */
#else
	vs_ctl_select,		/* select */
#endif	
	vs_ctl_ioctl,		/* ioctl */
	NULL,			/* mmap */
	vs_ctl_open,		/* open */
	vs_ctl_close		/* release */
};

/* ---------------------------------------------------------------------
 * /proc fs routines....
 */

#if (LINUX_VERSION_CODE >= 131394) /* Linux 2.1.66 */
static inline int line_info(char *buf, struct vs1000_state *vs, int line)
{
	struct vs1000_port_state	*pstate;
	struct nr_port			*info;
	char				stat_buf[40];
	int				ret;
	unsigned int			control, status;

	ret = sprintf(buf, "%d: ", line + vs->base_port);

	info = nrp_table[line + vs->base_port];
	pstate = vs->port_state + line;
	
	/*
	 * Figure out the current RS-232 lines
	 */
	status = pstate->modemStatusRegister;
	control = info ? info->modem_control_register : 0;
	stat_buf[0] = 0;
	stat_buf[1] = 0;
	if (control & MCR_DTR_SET_MASK)
		strcat(stat_buf, "|RTS");
	if (status & MSR_CTS_ON)
		strcat(stat_buf, "|CTS");
	if (control & MCR_DTR_SET_MASK)
		strcat(stat_buf, "|DTR");
	if (status & MSR_DSR_ON)
		strcat(stat_buf, "|DSR");
	if (status & MSR_CD_ON)
		strcat(stat_buf, "|CD");
	if (status & MSR_TX_FLOWED_OFF)
		strcat(stat_buf, "|TXOFF");

	if (info && info->baud_rate) {
		ret += sprintf(buf+ret, " baud:%d", info->baud_rate);
	}

#if 0
	ret += sprintf(buf+ret, " tx:%d rx:%d",
		      state->icount.tx, state->icount.rx);

	if (state->icount.frame)
		ret += sprintf(buf+ret, " fe:%d", state->icount.frame);
	
	if (state->icount.parity)
		ret += sprintf(buf+ret, " pe:%d", state->icount.parity);
	
	if (state->icount.brk)
		ret += sprintf(buf+ret, " brk:%d", state->icount.brk);	

	if (state->icount.overrun)
		ret += sprintf(buf+ret, " oe:%d", state->icount.overrun);
#endif

	/*
	 * Last thing is the RS-232 status lines
	 */
	ret += sprintf(buf+ret, " %s\n", stat_buf+1);
	return ret;
}

static int rs_read_proc(char *page, char **start, off_t off, int count,
		 int *eof, void *data)
{
	int i, j, len = 0, l;
	off_t	begin = 0;
	struct vs1000_state *vs;
	
	len += sprintf(page, "serinfo:1.0 driver:%s\n", VS_VERSION);
	for (i = 0; i < NUM_VS1000; i++) {
		vs = conc_state[i];
		if (!vs)
			continue;
		for (j=0; j < vs->num_ports; j++) {
			l = line_info(page + len, vs, j);
			len += l;
			if (len+begin > off+count)
				goto done;
			if (len+begin < off) {
				begin += len;
				len = 0;
			}
		}
	}
	*eof = 1;
done:
	if (off >= len+begin)
		return 0;
	*start = page + (begin-off);
	return ((count < begin+len-off) ? count : begin+len-off);
}
#endif

/*
 * ----------------------------------------------------------------
 * Initialization routines
 */

/*
 * 	int boxid, int num_lines, int base_port, 
 * 			     struct device *dev,
 * 			     char box_address[6])
 * 
 */

static int init_vs1000_state(struct vs1000_box_config *in, int from_user)
{
	struct vs1000_state *vs;
	struct device *dev;
	int i, ret;
	unsigned int	size;
	struct vs1000_box_config *cfg, tmp;

	if (from_user) {
		if (copy_from_user(&tmp, (void *) in,
				   sizeof(struct vs1000_box_config)))
			return -EFAULT;
		cfg = &tmp;
	} else {
		cfg = in;
	}

	if (cfg->boxid >= NUM_VS1000)
		return -EINVAL;

	dev = dev_get(cfg->interface);
	if (!dev)
		return -ENOENT;

	if (conc_state[cfg->boxid]) {
		ret = free_vs1000_box(cfg->boxid);
		if (ret)
			return ret;
		conc_state[cfg->boxid] = 0;
	}
	if (cfg->num_ports == 0)
		return 0;

	vs = kmalloc(sizeof(struct vs1000_state), GFP_KERNEL);
	if (!vs)
		return -ENOMEM;
	memset(vs, 0, sizeof(struct vs1000_state));
	memcpy(vs->ether_addr, cfg->ether_addr, 6);
	vs->dev = dev;
	skb_queue_head_init(&vs->rtq);
	vs->num_ports = cfg->num_ports;
	vs->pkt_window_size = 8;
	vs->timer.data = (unsigned long) vs;
	vs->timer.function = vs1000_timer;
	vs->base_port = 0;
#ifdef atomic_set
	atomic_set(&vs->timer_active, 1);
#else
	vs->timer_active = 1;
#endif
	for (i=0; i < cfg->boxid; i++) {
		if (conc_state[i])
			vs->base_port += conc_state[i]->num_ports;
	}

	size = cfg->num_ports * sizeof(struct vs1000_port_state);
	vs->port_state = kmalloc(size, GFP_KERNEL);
	if (!vs->port_state) {
		kfree(vs);
		return -ENOMEM;
	}
	memset(vs->port_state, 0, size);
	
	for (i = 0; i < cfg->num_ports; i++)
		init_nr_port(vs, i + vs->base_port);

	conc_state[cfg->boxid] = vs;
	return 0;
}

static int free_vs1000_box(int boxid)
{
	struct vs1000_state *vs;
	struct nr_port *info;
	int	i;

	if (boxid >= NUM_VS1000 || !conc_state[boxid])
		return -EINVAL;

	vs = conc_state[boxid];

	for (i=0; i < vs->num_ports; i++) {
		info = nrp_table[i + vs->base_port];
		if (!info)
			continue;
		if (info->count || info->blocked_open)
			return -EBUSY;
	}

	if (vs->timer.prev)
		del_timer(&vs->timer);

	conc_state[boxid] = 0;

	for (i=0; i < vs->num_ports; i++) {
		info = nrp_table[i + vs->base_port];
		if (info) {
			nrp_table[i + vs->base_port] = 0;
			kfree(info);
		}
	}

	if (vs->port_state)
		kfree(vs->port_state);
	kfree(vs);
	return 0;
}

static __initfunc(int vs1000_tty_init())
{
	int	retval;

	/*
	 * Initialize the array of pointers to our own internal state
	 * structures.
	 */
	memset(nrp_table, 0, sizeof(nrp_table));

	/*
	 * Set up the tty driver structure and then register this
	 * driver with the tty layer.
	 */
	memset(&netrocket_driver, 0, sizeof(struct tty_driver));
	netrocket_driver.magic = TTY_DRIVER_MAGIC;
#if (LINUX_VERSION_CODE > 131336) /* 2.1.8 (?) */
	netrocket_driver.driver_name = "VS-1000";
#endif
	netrocket_driver.name = "ttyV";
	netrocket_driver.major = TTY_NETROCKET_MAJOR;
	netrocket_driver.minor_start = 0;
	netrocket_driver.num = MAX_NRP_PORTS;
	netrocket_driver.type = TTY_DRIVER_TYPE_SERIAL;
	netrocket_driver.subtype = SERIAL_TYPE_NORMAL;
	netrocket_driver.init_termios = tty_std_termios;
	netrocket_driver.init_termios.c_cflag =
		B9600 | CS8 | CREAD | HUPCL | CLOCAL;
	netrocket_driver.flags = TTY_DRIVER_REAL_RAW;
	netrocket_driver.refcount = &netrocket_refcount;
	netrocket_driver.table = netrocket_table;
	netrocket_driver.termios = netrocket_termios;
	netrocket_driver.termios_locked = netrocket_termios_locked;

	netrocket_driver.open = nrp_open;
	netrocket_driver.close = nrp_close;
	netrocket_driver.write = nrp_write;
	netrocket_driver.put_char = nrp_put_char;
	netrocket_driver.flush_chars = nrp_flush_chars;
	netrocket_driver.write_room = nrp_write_room;
	netrocket_driver.chars_in_buffer = nrp_chars_in_buffer;
	netrocket_driver.flush_buffer = nrp_flush_buffer;
	netrocket_driver.ioctl = nrp_ioctl;
	netrocket_driver.throttle = nrp_throttle;
	netrocket_driver.unthrottle = nrp_unthrottle;
	netrocket_driver.set_termios = nrp_set_termios;
	netrocket_driver.stop = nrp_stop;
	netrocket_driver.start = nrp_start;
	netrocket_driver.hangup = nrp_hangup;
#if (LINUX_VERSION_CODE >= 131394) /* Linux 2.1.66 */
	netrocket_driver.break_ctl = nrp_break;
	netrocket_driver.read_proc = rs_read_proc;
#endif
#if (LINUX_VERSION_CODE >= 131343)
	netrocket_driver.send_xchar = nrp_send_xchar;
	netrocket_driver.wait_until_sent = nrp_wait_until_sent;
#endif

	/*
	 * The callout device is just like normal device except for
	 * the minor number and the subtype code.
	 */
	callout_driver = netrocket_driver;
	callout_driver.name = "cuV";
	callout_driver.major = CUA_NETROCKET_MAJOR;
	callout_driver.minor_start = 0;
	callout_driver.subtype = SERIAL_TYPE_CALLOUT;
#if (LINUX_VERSION_CODE >= 131394) /* Linux 2.1.66 */
	callout_driver.read_proc = 0;
	callout_driver.proc_entry = 0;
#endif
	
	retval = tty_register_driver(&callout_driver);
	if (retval < 0) {
		printk("Couldn't install Rocketport callout driver "
		       "(error %d)\n", -retval);
		return -1;
	}

	retval = tty_register_driver(&netrocket_driver);
	if (retval < 0) {
		printk("Couldn't install tty Rocketport driver "
		       "(error %d)\n", -retval);
		return -1;
	}
	return 0;
}	
	   

/* Called by proto.c on kernel start up */
static __initfunc(int vs1000_proto_init(struct net_proto *pro))
{
	comtrol_packet_type.type = htons(ETH_P_COMTROL);
	dev_add_pack(&comtrol_packet_type);

	memset(&conc_state, 0, sizeof(conc_state));
	wake_daemon_wait = NULL;
	return 0;
}

#ifdef MODULE

int init_module(void)
{
	int	retval;

	printk("VS-1000 driver, version %s (%s)", VS_VERSION, VS_DATE);

	retval = vs1000_tty_init();
	if (retval)
		return retval;
	control_major = register_chrdev(0, vs_ctl_name, &ctl_fops);
	if (control_major < 0) {
		printk("Failed to register the control registry.\n");
		return (control_major);
	}
	retval = vs1000_proto_init(NULL);
	return (retval);
}


void cleanup_module(void)
{
	int	retval, i;
	
	dev_remove_pack(&comtrol_packet_type);
	if (control_major > 0) {
		retval = unregister_chrdev(control_major, vs_ctl_name);
		if (retval) {
			printk("Error %d while trying to unregister "
			       "control device.\n", -retval);
		} else
			control_major = 0;
	}
	retval = tty_unregister_driver(&callout_driver);
	if (retval) {
		printk("Error %d while trying to unregister "
		       "rocketport callout driver\n", -retval);
	}
	retval = tty_unregister_driver(&netrocket_driver);
	if (retval) {
		printk("Error %d while trying to unregister "
		       "rocketport driver\n", -retval);
	}
	if (tmp_buf)
		free_page((unsigned long) tmp_buf);
	
	for (i=0; i < NUM_VS1000; i++)
		free_vs1000_box(i);
	
	return;
}

#endif  /* MODULE */
