/*
 * vs1000d --- vs1000 support daemon
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <poll.h>
#include <ctype.h>
#include <string.h>
#include <signal.h>
#include <syslog.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include "vs1000.h"
#include "vs1000d.h"
#include "version.h"

char *device_name = "/dev/ttyVC";
char *dfl_boot_file = "/etc/vs1000.bin";
const char *not_loaded = "VS-1000 driver not loaded?";
const char *pid_file = "/var/run/vs1000d.pid";

static int get_major_number(const char *name)
{
	FILE	*f;
	char	buf[80], devname[80];
	int	n, major = -1;

	f = fopen("/proc/devices", "r");
	if (!f)
		return -1;
	while (!feof(f)) {
		if (!fgets(buf, sizeof(buf), f))
			break;
		if (sscanf(buf, "%d %s", &n, devname) != 2)
			continue;
		if (!strcmp(devname, name)) {
			major = n;
			break;
		}
	}
	fclose(f);
	return major;
}

static int open_dev(dev_t dev, int mode)
{
	char *fn;
	int fd;

	if ((fn = tempnam(NULL, "vsdev")) == NULL)
		return -1;
	if (mknod(fn, S_IFCHR | 0600, dev) != 0)
		return -1;
	fd = open(fn, mode);
	unlink(fn);
	return fd;
}

int config_vs1000(int fd, int boxid, char *emsg)
{
	struct vs_box_config *config;
	struct vs1000_box_config cfg;
	int ret;

	if (boxid >= num_box_config)
		return -1;

	config = box_config + boxid;

	memset(&cfg, 0, sizeof(cfg));
	cfg.boxid = boxid;
	cfg.num_ports = config->num_lines;
	memcpy(cfg.ether_addr, config->ether_addr, 6);
	strcpy(cfg.interface, config->interface);
	ret = ioctl(fd, VS1000_SET_BOX_CONFIG, &cfg);
	if (ret) {
		sprintf(emsg, "Couldn't configure VS-1000 box #%d", boxid);
		return -1;
	}
	return 0;
}

#define START_OFFSET	0x00010000UL
#define GO_OFFSET	0x10000000UL

int boot_vs1000(int fd, int boxid, const char *boot_file, char *emsg)
{
	char	buf[1000];
	int	bfd;
	struct vs1000_load_struct vload;
	int	size, ret;
	const char *whilst = "while booting VS-1000 box #";

	bfd = open(boot_file, O_RDONLY);
	if (bfd < 0) {
		sprintf(emsg, "Couldn't open boot file %s %s%d",
			boot_file, whilst, boxid);
		return -1;
	}
	vload.boxid = boxid;
	vload.offset = START_OFFSET;
	vload.buf = buf;
	while (1) {
		size = read(bfd, buf, sizeof(buf));
		if (size == 0)
			break;
		if (size < 0) {
			if (errno == EINTR)
				continue;
			sprintf(emsg, "Error reading boot file %s %s%d",
				boot_file, whilst, boxid);
			close(bfd);
			return -1;
		}
		vload.size = size;
		ret = ioctl(fd, VS1000_LOAD_CODE, &vload);
		if (ret) {
			sprintf(emsg, "Error loading microcode %s%d",
				whilst, boxid);
			close(bfd);
			return -1;
		}
		vload.offset += size;
	}
	close(bfd);

	vload.size = 0;
	vload.offset = GO_OFFSET;
	vload.buf = 0;
	ret = ioctl(fd, VS1000_LOAD_CODE, &vload);
	if (ret) {
		sprintf(emsg, "Couldn't send go command %s %d",
			whilst, boxid);
		return -1;
	}
	return 0;
}

int reset_vs1000(int fd, int boxid, char *emsg)
{
	struct vs1000_load_struct vload;
	int	ret;

	vload.boxid = boxid;
	vload.offset = 0;
	vload.size = 0xFFFFFFFF;
	vload.buf = 0;
	ret = ioctl(fd, VS1000_LOAD_CODE, &vload);
	if (ret) {
		sprintf(emsg, "Couldn't reset VS-1000 box #%d", boxid);
		return -1;
	}
	return 0;
}

/*
 * Configure and boot the boxes
 */	
int initial_load_boxes(int fd, const char *boot_file)
{
	int	i;
	char	emsg[256];
	int	dead_boxes = 0;

	for (i=0; i < num_box_config; i++) {
		if (config_vs1000(fd, i, emsg) < 0) {
			syslog(LOG_ERR, "%s", emsg);
			fprintf(stderr, "%s\n", emsg);
			exit(1);
		}
		if (reset_vs1000(fd, i, emsg) < 0) {
			syslog(LOG_ERR, "%s", emsg);
			fprintf(stderr, "%s\n", emsg);
			exit(1);
		}
	}
	sleep(5);
	for (i=0; i < num_box_config; i++) {
		if (boot_vs1000(fd, i, boot_file, emsg) < 0) {
			syslog(LOG_ERR, "%s", emsg);
			fprintf(stderr, "%s\n", emsg);
			dead_boxes |= 1 << i;
		}
	}
	return dead_boxes;
}

int service_timeout = -1;
#define INITIAL_TIMEOUT	30 * 1000 /* 30 seconds */
#define MAX_TIMEOUT	60 * 60 * 1000 /* 1 hour */

void service_devices(int fd, char *boot_file, int initial_dead)
{
	struct pollfd 	poll_ctl;
	int		i, dead_boxes, status, ret;
	char		emsg[256];

	printf("In service device...\n");
	poll_ctl.fd = fd;
	poll_ctl.events = POLLOUT;
	poll_ctl.revents = 0;
	dead_boxes = initial_dead;
	service_timeout = dead_boxes ? 0 : -1;
	while (1) {
		if ((ret = poll(&poll_ctl, 1, service_timeout)) < 0) {
			if (errno == EINTR)
				continue;
			syslog(LOG_ERR, "poll: %m");
			exit(1);
		}
		if (ioctl(fd, VS1000_GET_PROBLEM_BOXES, &status)) {
			syslog(LOG_ERR, "VS1000_GET_PROBLEM_BOXES: %m");
			exit(1);
		}
		printf("Woke up!  Problem status code %d\n", status);
		status |= dead_boxes;
		for (i=0; i < num_box_config; i++) {
			if ((status & (1 << i)) == 0)
				continue;
			if (boot_vs1000(fd, i, boot_file, emsg) < 0) {
				syslog(LOG_WARNING, "%s", emsg);
				dead_boxes |= 1 << i;
				if (service_timeout < INITIAL_TIMEOUT)
					service_timeout = INITIAL_TIMEOUT;
				else if (service_timeout < MAX_TIMEOUT)
					service_timeout = 2 * service_timeout;
			} else {
				syslog(LOG_INFO,
				       "Successfully rebooted box %d", i);
				dead_boxes &= ~(1 << i);
			}
		}
		if (!dead_boxes)
			service_timeout = -1;
	}
}

int become_daemon()
{
	int child_pid, fd;

	child_pid = fork();
	if (child_pid == -1)
		return -1;
	if (child_pid)
		exit(0);
	(void) chdir("/");
	fd = open("/dev/null", O_RDWR, 0);
	if (fd >= 0) {
		(void) dup2(fd, 0);
		(void) dup2(fd, 1);
		(void) dup2(fd, 2);
		if (fd > 2)
			(void) close(fd);
	}
	return (0);
}

void check_pid_file()
{
	FILE	*f;
	int	pid;
	
	if (access(pid_file, R_OK) == 0) {
		f = fopen(pid_file, "r");
		if (f) {
			pid = 0;
			fscanf(f, "%d", &pid);
			if (pid && !kill(pid, 0)) {
				fprintf(stderr,
					"vs1000 daemon already running!\n");
				syslog(LOG_ERR,
				       "vs1000 daemon already running");
				exit(1);
			}
			fclose(f);
		}
	}
}

void put_pid_file()
{
	FILE	*f;

	f = fopen(pid_file, "w");
	if (!f) {
		syslog(LOG_WARNING, "Warning: couldn't write pid file %s",
		       pid_file);
		return;
	}
	fprintf(f, "%d\n", getpid());
	fclose(f);
}

/*
 * Create a device if it doesn't exist and have the correct device
 * number.
 */
#define MKDEV(a,b) ((int)((((a) & 0xff) << 8) | ((b) & 0xff)))

void create_device(const char *prefix, int major, int num, gid_t def_group)
{
	char	name[80];
	struct stat	st;
	dev_t		device;

	sprintf(name, "%s%d", prefix, num);
	device = MKDEV(major, num);
	if (lstat(name, &st) == 0) {
		if ((st.st_rdev == device) &&
		    S_ISCHR(st.st_mode))
			return;
#if 0
		syslog(LOG_INFO, "%s has dev %d, mode %o wanted %d",
		       name, st.st_rdev, st.st_mode, device);
#endif
		(void) unlink(name);
	}
	if (mknod(name, S_IFCHR | 0620, device) < 0) {
		perror(name);
		syslog(LOG_ERR, "Couldn't create device %s: %m", name);
	}
	if (chown(name, 0, def_group) < 0) {
		perror(name);
		syslog(LOG_ERR, "Couldn't chmod device %s: %m", name);
	}
}

/*
 * This function creates all of the devices
 */
void create_devices()
{
	int	i, num_ports;
	struct group *grp;
	gid_t	tty_group = 0;
	int major, callout_major;

	major = get_major_number("ttyV");
	callout_major = get_major_number("cuV");

	if (major == 0 && callout_major == 0)
		return;

	num_ports = 0;
	for (i=0; i < num_box_config; i++)
		num_ports += box_config[i].num_lines;

	grp = getgrnam("tty");
	if (grp)
		tty_group = grp->gr_gid;

	for (i=0; i < num_ports; i++) {
		if (major)
			create_device("/dev/ttyV", major, i, tty_group);
		if (callout_major)
			create_device("/dev/cuV", callout_major, i, tty_group);
	}
}


void sigterm_handler()
{
	syslog(LOG_NOTICE, "Exiting");
	(void) unlink(pid_file);
	exit(0);
}

void sighup_handler()
{
	service_timeout = 0;
	signal(SIGHUP, sighup_handler);
	return;
}


int main(int argc, char **argv)
{
	int	major, fd, dead_boxes;

	if (getuid()) {
		fprintf(stderr, "You must be root to run vs1000d!\n");
		exit(1);
	}
	
	openlog("vs1000d", LOG_NDELAY|LOG_PID, LOG_DAEMON);
	parse_config("/etc/vs1000.conf");
	if (!boot_file)
		boot_file = dfl_boot_file;
	
	major = get_major_number("VS1000ctl");
	if (major < 0) {
		syslog(LOG_ERR, "Couldn't load control file");
		fprintf(stderr, "Couldn't find control file.  %s\n",
			not_loaded);
		exit(1);
	}
	fd = open_dev(major << 8, O_RDWR);
	if (fd < 0) {
		syslog(LOG_ERR, "Couldn't open control file");
		fprintf(stderr, "Couldn't open control device.  %s\n",
			not_loaded);
		exit(1);
	}
	check_pid_file();
	if (become_daemon() < 0) {
		syslog(LOG_ERR, "Couldn't become daemon");
		fprintf(stderr, "Error forking child process while "
			"starting daemon mode\n");
		exit(1);
	}
	put_pid_file();
	syslog(LOG_INFO, "Starting VS-1000 daemon, version %s (%s)",
	       VS_VERSION, VS_DATE);
	signal(SIGHUP, sighup_handler);
	signal(SIGTERM, sigterm_handler);
	signal(SIGINT, sigterm_handler);
	create_devices();
	dead_boxes = initial_load_boxes(fd, boot_file);
	service_devices(fd, boot_file, dead_boxes);
	return 0;
}
